###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################

# Import Pyomo libraries
from pyomo.environ import (Var,
                           Set,
                           NonNegativeReals,
                           NegativeReals,
                           Reference,
                           units as pyunits,
                           exp,
                           value,
                           check_optimal_termination)
# Import IDAES cores
from idaes.core import (ControlVolume0DBlock,
                        declare_process_block_class)
from idaes.core.util.exceptions import ConfigurationError
from idaes.core.util import get_solver
from idaes.core.util.misc import add_object_reference
import idaes.core.util.scaling as iscale
from watertap.core.util.initialization import check_solve, check_dof
from watertap.unit_models._reverse_osmosis_base import (ConcentrationPolarizationType,
        MassTransferCoefficient,
        PressureChangeType,
        _ReverseOsmosisBaseData)
import idaes.logger as idaeslog


__author__ = "Tim Bartholomew, Adam Atia"

# Set up logger
_log = idaeslog.getLogger(__name__)


@declare_process_block_class("ReverseOsmosis0D")
class ReverseOsmosisData(_ReverseOsmosisBaseData):
    """
    Standard RO Unit Model Class:
    - zero dimensional model
    - steady state only
    - single liquid phase only
    """
    CONFIG = _ReverseOsmosisBaseData.CONFIG()

    def build(self):
        """
        Build the RO model.
        """
        # Call UnitModel.build to setup dynamics
        super().build()

        # for quacking like 1D model -> 0. is "in", 1. is "out"
        self.length_domain = Set(ordered=True, initialize=(0., 1.))  # inlet/outlet set
        add_object_reference(self, 'difference_elements', self.length_domain)
        self.first_element = self.length_domain.first()

        # Build control volume for feed side
        self.feed_side = ControlVolume0DBlock(default={
            "dynamic": False,
            "has_holdup": False,
            "property_package": self.config.property_package,
            "property_package_args": self.config.property_package_args})

        self.feed_side.add_state_blocks(
            has_phase_equilibrium=False)

        self.feed_side.add_material_balances(
            balance_type=self.config.material_balance_type,
            has_mass_transfer=True)

        self.feed_side.add_energy_balances(
            balance_type=self.config.energy_balance_type,
            has_enthalpy_transfer=True)

        self.feed_side.add_momentum_balances(
            balance_type=self.config.momentum_balance_type,
            has_pressure_change=self.config.has_pressure_change)

        # for quacking like 1D model
        add_object_reference(self.feed_side,'properties',
                {**{(t,0.) : self.feed_side.properties_in[t] for t in self.flowsheet().config.time},
                 **{(t,1.) : self.feed_side.properties_out[t] for t in self.flowsheet().config.time}}
                )

        # Add additional state blocks
        tmp_dict = dict(**self.config.property_package_args)
        tmp_dict["has_phase_equilibrium"] = False
        tmp_dict["parameters"] = self.config.property_package
        tmp_dict["defined_state"] = False  # these blocks are not inlets
        # Build permeate side
        self.permeate_side = self.config.property_package.state_block_class(
            self.flowsheet().config.time,
            self.length_domain,
            doc="Material properties of permeate along permeate channel",
            default=tmp_dict)
        self.mixed_permeate = self.config.property_package.state_block_class(
            self.flowsheet().config.time,
            doc="Material properties of mixed permeate exiting the module",
            default=tmp_dict)
        # Interface properties
        self.feed_side.properties_interface = self.config.property_package.state_block_class(
            self.flowsheet().config.time,
            self.length_domain,
            doc="Material properties of feed-side membrane interface",
            default=tmp_dict)

        # Add Ports
        self.add_inlet_port(name='inlet', block=self.feed_side)
        self.add_outlet_port(name='retentate', block=self.feed_side)
        self.add_port(name='permeate', block=self.mixed_permeate)

        # References for control volume
        # pressure change
        if (self.config.has_pressure_change and
                self.config.momentum_balance_type != 'none'):
            self.deltaP = Reference(self.feed_side.deltaP)

        self._make_performance()

        self._add_expressions()


    def _make_performance(self):

        units_meta = self.config.property_package.get_metadata().get_derived_units

        solvent_set = self.config.property_package.solvent_set
        solute_set = self.config.property_package.solute_set

        if self.config.pressure_change_type == PressureChangeType.calculated:
            self.dP_dx = Var(
                self.flowsheet().config.time,
                self.length_domain,
                initialize=-5e4,
                bounds=(-2e5, -1e3),
                domain=NegativeReals,
                units=units_meta('pressure')*units_meta('length')**-1,
                doc="Pressure drop per unit length of feed channel at inlet and outlet")
        elif self.config.pressure_change_type == PressureChangeType.fixed_per_unit_length:
            self.dP_dx = Var(
                self.flowsheet().config.time,
                initialize=-5e4,
                bounds=(-2e5, -1e3),
                domain=NegativeReals,
                units=units_meta('pressure')*units_meta('length')**-1,
                doc="pressure drop per unit length across feed channel")

        if ((self.config.pressure_change_type != PressureChangeType.fixed_per_stage)
                or (self.config.mass_transfer_coefficient == MassTransferCoefficient.calculated)):
            # comes from ControlVolume1D in 1DRO
            self.length = Var(
                initialize=10,
                bounds=(0.1, 5e2),
                domain=NonNegativeReals,
                units=units_meta('length'),
                doc='Effective membrane length')
            # not optional in 1DRO
            self.width = Var(
                initialize=1,
                bounds=(0.1, 5e2),
                domain=NonNegativeReals,
                units=units_meta('length'),
                doc='Effective feed-channel width')

        if (self.config.mass_transfer_coefficient == MassTransferCoefficient.calculated
            or self.config.pressure_change_type == PressureChangeType.calculated):
            self.area_cross = Var(
                    initialize=1e-3*1*0.95,
                    bounds=(0, 1e3),
                    domain=NonNegativeReals,
                    units=units_meta('length')**2,
                    doc='Cross sectional area')

        super()._make_performance()

        # mass transfer
        def mass_transfer_phase_comp_initialize(b, t, p, j):
            return value(self.feed_side.properties_in[t].get_material_flow_terms('Liq', j)
                         * self.recovery_mass_phase_comp[t, 'Liq', j])
        self.mass_transfer_phase_comp = Var(
            self.flowsheet().config.time,
            self.config.property_package.phase_list,
            self.config.property_package.component_list,
            initialize=mass_transfer_phase_comp_initialize,
            bounds=(1e-8, 1e6),
            domain=NonNegativeReals,
            units=units_meta('mass') * units_meta('time')**-1,
            doc='Mass transfer to permeate')

        # constraints for additional variables (i.e. variables not used in other constraints)

        @self.Constraint(self.flowsheet().config.time,
                         self.config.property_package.phase_list,
                         self.config.property_package.component_list,
                         doc="Mass transfer term")
        def eq_mass_transfer_term(self, t, p, j):
            return self.mass_transfer_phase_comp[t, p, j] == -self.feed_side.mass_transfer_term[t, p, j]

        # Different expression in 1DRO
        @self.Constraint(self.flowsheet().config.time,
                         self.config.property_package.phase_list,
                         self.config.property_package.component_list,
                         doc="Permeate production")
        def eq_permeate_production(b, t, p, j):
            return (b.mixed_permeate[t].get_material_flow_terms(p, j)
                    == b.area * b.flux_mass_phase_comp_avg[t, p, j])

        # Feed and permeate-side connection
        @self.Constraint(self.flowsheet().config.time,
                         self.config.property_package.phase_list,
                         self.config.property_package.component_list,
                         doc="Mass transfer from feed to permeate")
        def eq_connect_mass_transfer(b, t, p, j):
            return (b.mixed_permeate[t].get_material_flow_terms(p, j)
                    == -b.feed_side.mass_transfer_term[t, p, j])

        # Non-existent in 1DRO
        @self.Constraint(self.flowsheet().config.time,
                         doc="Enthalpy transfer from feed to permeate")
        def eq_connect_enthalpy_transfer(b, t):
            return (b.mixed_permeate[t].get_enthalpy_flow_terms('Liq')
                    == -b.feed_side.enthalpy_transfer[t])

        # # Permeate-side stateblocks
        # Not in 1DRO
        @self.Constraint(self.flowsheet().config.time,
                         self.length_domain,
                         solute_set,
                         doc="Permeate mass fraction")
        def eq_mass_frac_permeate(b, t, x, j):
            return (b.permeate_side[t, x].mass_frac_phase_comp['Liq', j]
                    * sum(self.flux_mass_phase_comp[t, x, 'Liq', jj]
                          for jj in self.config.property_package.component_list)
                    == self.flux_mass_phase_comp[t, x, 'Liq', j])

        # not in 1DRO
        @self.Constraint(self.flowsheet().config.time,
                         self.length_domain,
                         doc="Permeate flowrate")
        def eq_flow_vol_permeate(b, t, x):
            return b.permeate_side[t, x].flow_vol_phase['Liq'] == b.mixed_permeate[t].flow_vol_phase['Liq']

        if self.config.pressure_change_type == PressureChangeType.fixed_per_unit_length:
            # Pressure change equation when dP/dx = user-specified constant,
            @self.Constraint(self.flowsheet().config.time,
                             doc="pressure change due to friction")
            def eq_pressure_change(b, t):
                return b.deltaP[t] == b.dP_dx[t] * b.length

        elif self.config.pressure_change_type == PressureChangeType.calculated:

            # Average pressure change per unit length due to friction
            @self.Expression(self.flowsheet().config.time,
                             doc="expression for average pressure change per unit length due to friction")
            def dP_dx_avg(b, t):
                return 0.5 * sum(b.dP_dx[t, x] for x in b.length_domain)

            # Pressure change equation
            @self.Constraint(self.flowsheet().config.time,
                             doc="pressure change due to friction")
            def eq_pressure_change(b, t):
                return b.deltaP[t] == b.dP_dx_avg[t] * b.length


    def _add_expressions(self):
        super()._add_expressions()

        @self.Expression(self.flowsheet().config.time,
                         doc='Over pressure ratio')
        def over_pressure_ratio(b, t):
            return (b.feed_side.properties_out[t].pressure_osm
                    - b.permeate_side[t,1.].pressure_osm) / \
                    b.feed_side.properties_out[t].pressure

    def initialize_build(blk,
                   initialize_guess=None,
                   state_args=None,
                   outlvl=idaeslog.NOTSET,
                   solver=None,
                   optarg=None,
                   fail_on_warning=False,
                   ignore_dof=False):
        """
        General wrapper for RO initialization routines

        Keyword Arguments:

            initialize_guess : a dict of guesses for solvent_recovery, solute_recovery,
                               and cp_modulus. These guesses offset the initial values
                               for the retentate, permeate, and membrane interface
                               state blocks from the inlet feed
                               (default =
                               {'deltaP': -1e4,
                               'solvent_recovery': 0.5,
                               'solute_recovery': 0.01,
                               'cp_modulus': 1.1})
            state_args : a dict of arguments to be passed to the property
                         package(s) to provide an initial state for the inlet
                         feed side state block (see documentation of the specific
                         property package) (default = None).
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None)
            solver : solver object or string indicating which solver to use during
                     initialization, if None provided the default solver will be used
                     (default = None)
            fail_on_warning : boolean argument to fail or only produce  warning upon unsuccessful solve (default=False)
            ignore_dof : boolean argument to ignore when DOF != 0 (default=False)
        Returns:
            None
        """

        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(blk.name, outlvl, tag="unit")
        # Set solver and options
        opt = get_solver(solver, optarg)


        # ---------------------------------------------------------------------
        # Extract initial state of inlet feed
        source = blk.feed_side.properties_in[blk.flowsheet().config.time.first()]
        state_args = blk._get_state_args(source, blk.mixed_permeate[0], initialize_guess, state_args)

        # Initialize feed inlet state block
        flags_feed_side = blk.feed_side.properties_in.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args['feed_side'],
            hold_state=True)

        init_log.info("Initialization Step 1 Complete.")
        if not ignore_dof:
            check_dof(blk, fail_flag=fail_on_warning, logger=init_log)
        # ---------------------------------------------------------------------
        # Initialize other state blocks
        # base properties on inlet state block

        blk.feed_side.properties_out.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args['retentate'],)
        blk.feed_side.properties_interface.initialize(
                outlvl=outlvl,
                optarg=optarg,
                solver=solver,
                state_args=state_args['interface'],)
        blk.mixed_permeate.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args['permeate'],)
        blk.permeate_side.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args['permeate'],)
        init_log.info("Initialization Step 2 Complete.")

        # ---------------------------------------------------------------------
        # Solve unit
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(blk, tee=slc.tee)
            # occasionally it might be worth retrying a solve
            if not check_optimal_termination(res):
                init_log.warn("Trouble solving ReverseOsmosis0D unit model, trying one more time")
                res = opt.solve(blk, tee=slc.tee)
        check_solve(res, checkpoint='Initialization Step 3', logger=init_log, fail_flag=fail_on_warning)
        # ---------------------------------------------------------------------
        # Release Inlet state
        blk.feed_side.release_state(flags_feed_side, outlvl)
        init_log.info(
            "Initialization Complete: {}".format(idaeslog.condition(res))
        )

    def calculate_scaling_factors(self):
        # setting scaling factors for variables
        # will not override if the user does provide the scaling factor
        if iscale.get_scaling_factor(self.dens_solvent) is None:
            sf = iscale.get_scaling_factor(self.feed_side.properties_in[0].dens_mass_phase['Liq'])
            iscale.set_scaling_factor(self.dens_solvent, sf)

        super().calculate_scaling_factors()

        for (t, p, j), v in self.mass_transfer_phase_comp.items():
            sf = iscale.get_scaling_factor(self.feed_side.properties_in[t].get_material_flow_terms(p, j))
            if iscale.get_scaling_factor(v) is None:
                iscale.set_scaling_factor(v, sf)
            v = self.feed_side.mass_transfer_term[t,p,j]
            if iscale.get_scaling_factor(v) is None:
                iscale.set_scaling_factor(v, sf)

        if hasattr(self, 'area_cross'):
            if iscale.get_scaling_factor(self.area_cross) is None:
                iscale.set_scaling_factor(self.area_cross, 100)

        if hasattr(self, 'length'):
            if iscale.get_scaling_factor(self.length) is None:
                iscale.set_scaling_factor(self.length, 1)

        if hasattr(self, 'width'):
            if iscale.get_scaling_factor(self.width) is None:
                iscale.set_scaling_factor(self.width, 1)

        if hasattr(self, 'dP_dx'):
            for v in self.dP_dx.values():
                if iscale.get_scaling_factor(v) is None:
                    iscale.set_scaling_factor(v, 1e-4)



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################

"""
    This is a set of scaling utility helper functions used by the chemistry tests.
"""
from idaes.core.util import scaling as iscale
from pyomo.environ import value

__author__ = "Austin Ladshaw"

## Helper function for setting eps values associated with solubility_product functions
# NOTE: Function does nothing if no solubility reactions are present
def _set_eps_vals(rxn_params, rxn_config, factor=1e-2, max_k_eq_ref=1e-16):
    # For solubility reactions, have to set the eps value
    if (hasattr(rxn_params,"equilibrium_reaction_idx")):
        for rid in rxn_params.equilibrium_reaction_idx:
            # Grab the 'k_eq_ref' value from the reaction config
            scale = rxn_config["equilibrium_reactions"][rid]["parameter_data"]["k_eq_ref"][0]

            # NOTE: ONLY certain functions have an eps value that we need to set
            if (hasattr(rxn_params.component("reaction_"+rid),"eps")):
                # highest allowable value for setting eps based on k_eq_ref
                rxn_params.component("reaction_"+rid).eps.value = min(max_k_eq_ref, scale)*factor

## Helper function for setting scaling factors for equilibrium reactions
def _set_equ_rxn_scaling(unit, rxn_config, min_k_eq_ref=1e-3):
    #Add scaling factors for reactions (changes depending on if it is a log form or not)
    for i in unit.control_volume.equilibrium_reaction_extent_index:
        # i[0] = time, i[1] = reaction
        # Grab the 'k_eq_ref' value from the reaction config
        scale = max(min_k_eq_ref, rxn_config["equilibrium_reactions"][i[1]]["parameter_data"]["k_eq_ref"][0])
        iscale.set_scaling_factor(unit.control_volume.equilibrium_reaction_extent[0.0,i[1]], 10/scale)
        iscale.constraint_scaling_transform(
            unit.control_volume.reactions[0.0].equilibrium_constraint[i[1]], 0.1)

## Helper function for setting scaling factors for inherent reactions
def _set_inherent_rxn_scaling(unit, thermo_config, min_k_eq_ref=1e-3):
    #Add scaling factors for reactions (changes depending on if it is a log form or not)
    for i in unit.control_volume.inherent_reaction_extent_index:
        # i[0] = time, i[1] = reaction
        # Grab the 'k_eq_ref' value from the thermo config
        scale = max(min_k_eq_ref, thermo_config["inherent_reactions"][i[1]]["parameter_data"]["k_eq_ref"][0])
        iscale.set_scaling_factor(unit.control_volume.inherent_reaction_extent[0.0,i[1]], 10/scale)
        iscale.constraint_scaling_transform(unit.control_volume.properties_out[0.0].
                inherent_equilibrium_constraint[i[1]], 0.1)

## Helper function for setting scaling factors for rate reactions
## TODO: Need to work out better scaling for rate reactions using min_scale
def _set_rate_rxn_scaling(rxn_params, unit, min_scale=1e-3):
    for i in rxn_params.rate_reaction_idx:
        scale = value(unit.control_volume.reactions[0.0].reaction_rate[i].expr)
        iscale.set_scaling_factor(unit.control_volume.rate_reaction_extent[0.0,i], 1000/scale)

## Helper function for setting scaling factors for the mass balance for FpcTP state vars
def _set_mat_bal_scaling_FpcTP(unit, min_flow_mol_phase_comp=1e-3):
    # For species
    for i in unit.control_volume.properties_out[0.0].mole_frac_phase_comp:
        # i[0] = phase, i[1] = species
        scale = max(min_flow_mol_phase_comp, unit.inlet.flow_mol_phase_comp[0, i[0], i[1]].value)

        iscale.set_scaling_factor(unit.control_volume.properties_out[0.0].mole_frac_comp[i[1]], 10/scale)
        iscale.set_scaling_factor(unit.control_volume.properties_out[0.0].mole_frac_phase_comp[i], 10/scale)
        iscale.set_scaling_factor(unit.control_volume.properties_out[0.0].flow_mol_phase_comp[i], 10/scale)
        iscale.constraint_scaling_transform(unit.control_volume.material_balances[0.0,i[1]], 10/scale)

    if (hasattr(unit.control_volume,"volume")):
        iscale.set_scaling_factor(unit.control_volume.volume, 10/unit.volume[0.0].value)

## Helper function for setting scaling factors for the mass balance for FTPx state vars
def _set_mat_bal_scaling_FTPx(unit, min_mole_frac_comp=1e-3):
    # For species
    for i in unit.control_volume.properties_out[0.0].mole_frac_phase_comp:
        # i[0] = phase, i[1] = species
        scale = max(min_mole_frac_comp, unit.inlet.mole_frac_comp[0, i[1]].value)
        iscale.set_scaling_factor(unit.control_volume.properties_out[0.0].mole_frac_comp[i[1]], 10/scale)
        iscale.set_scaling_factor(unit.control_volume.properties_out[0.0].mole_frac_phase_comp[i], 10/scale)
        iscale.set_scaling_factor(unit.control_volume.properties_out[0.0].flow_mol_phase_comp[i], 10/scale)
        iscale.constraint_scaling_transform(
            unit.control_volume.properties_out[0.0].component_flow_balances[i[1]], 10/scale)
        iscale.constraint_scaling_transform(unit.control_volume.material_balances[0.0,i[1]], 10/scale)

    if (hasattr(unit.control_volume,"volume")):
        iscale.set_scaling_factor(unit.control_volume.volume, 10/unit.volume[0.0].value)

## Helper function for setting energy balance scaling factors
def _set_ene_bal_scaling(unit):
    max_enth_mol_phase = 1
    min_scale = 1
    for phase in unit.control_volume.properties_in[0.0].enth_mol_phase:
        val = max(min_scale, abs(value(unit.control_volume.properties_in[0.0].enth_mol_phase[phase].expr)))
        max_enth_mol_phase = max(val, max_enth_mol_phase)
        iscale.set_scaling_factor(unit.control_volume.properties_in[0.0]._enthalpy_flow_term[phase], 10/val)
        iscale.set_scaling_factor(unit.control_volume.properties_out[0.0]._enthalpy_flow_term[phase], 10/val)

    iscale.constraint_scaling_transform(unit.control_volume.enthalpy_balances[0.0], 10/max_enth_mol_phase)



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
This module contains the base class for all zero order unit models.
"""
from idaes.core import UnitModelBlockData, useDefault, declare_process_block_class
from idaes.core.util.config import is_physical_parameter_block
import idaes.logger as idaeslog
from idaes.core.util.exceptions import ConfigurationError
from idaes.core.util.tables import create_stream_table_dataframe

from pyomo.common.config import ConfigBlock, ConfigValue, In
from pyomo.environ import units as pyunits


# Some more inforation about this module
__author__ = "Andrew Lee"

# Set up logger
_log = idaeslog.getLogger(__name__)


@declare_process_block_class("ZeroOrderBase")
class ZeroOrderBaseData(UnitModelBlockData):
    """
    Standard base class for zero order unit models.

    This class contains the basic consistency checks and common methods for
    zero order type models.
    """

    CONFIG = ConfigBlock()
    CONFIG.declare('dynamic', ConfigValue(
        domain=In([False]),
        default=False,
        description='Dynamic model flag - must be False',
        doc='''All zero-order models are steady-state only'''))
    CONFIG.declare('has_holdup', ConfigValue(
        default=False,
        domain=In([False]),
        description='Holdup construction flag - must be False',
        doc='''Zero order models do not include holdup'''))
    CONFIG.declare('property_package', ConfigValue(
        default=useDefault,
        domain=is_physical_parameter_block,
        description='Property package to use for control volume',
        doc='''Property parameter object used to define property  calculations,
        **default** - useDefault.
        **Valid values:** {
        **useDefault** - use default package from parent model or flowsheet,
        **PhysicalParameterObject** - a PhysicalParameterBlock object.}'''))
    CONFIG.declare('property_package_args', ConfigBlock(
        implicit=True,
        description='Arguments to use for constructing property packages',
        doc='''A ConfigBlock with arguments to be passed to a property block(s)
        and used when constructing these, **default** - None.
        **Valid values:** {see property package for documentation.}'''))
    CONFIG.declare('database', ConfigValue(
        description='An instance of a WaterTAP Database to use for parameters.'
        ))
    CONFIG.declare('process_subtype', ConfigValue(
        description=
        'Process subtype to use when looking up parameters from database.'))

    def build(self):
        super().build()

        # Set a placeholder attributes
        # Placeholder for technology type string
        self._tech_type = None

        # Attribute indicating what parameters need to be fixed in the model
        self._has_recovery_removal = False
        self._fixed_perf_vars = []

        # Place holders for assigning methods
        self._initialize = None  # used to link to initization routine
        self._scaling = None  # used to link to scaling routine
        self._get_Q = None  # used to provide inlet volumetric flow

        # Attributed for storing contents of reporting output
        self._stream_table_dict = {}
        self._perf_var_dict = {}

        # Check that property package meets requirements
        if self.config.property_package.phase_list != ["Liq"]:
            raise ConfigurationError(
                f"{self.name} configured with invalid property package. "
                f"Zero-order models only support property packages with a "
                f"single phase named 'Liq'.")
        if (not hasattr(self.config.property_package, "solvent_set") or
                self.config.property_package.solvent_set != ["H2O"]):
            raise ConfigurationError(
                f"{self.name} configured with invalid property package. "
                f"Zero-order models only support property packages which "
                f"include 'H2O' as the only Solvent.")
        if not hasattr(self.config.property_package, "solute_set"):
            raise ConfigurationError(
                f"{self.name} configured with invalid property package. "
                f"Zero-order models require property packages to declare all "
                f"dissolved species as Solutes.")
        if (len(self.config.property_package.solute_set) !=
                len(self.config.property_package.component_list)-1):
            raise ConfigurationError(
                f"{self.name} configured with invalid property package. "
                f"Zero-order models only support `H2O` as a solvent and all "
                f"other species as Solutes.")

    def initialize_build(self, state_args=None, outlvl=idaeslog.NOTSET,
                         solver=None, optarg=None):
        '''
        Placeholder initialization routine, raises NotImplementedError
        '''
        if self._initialize is None or not callable(self._initialize):
            raise NotImplementedError()
        else:
            self._initialize(self, state_args=None, outlvl=idaeslog.NOTSET,
                             solver=None, optarg=None)

    def calculate_scaling_factors(self):
        '''
        Placeholder scaling routine, should be overloaded by derived classes
        '''
        super().calculate_scaling_factors()

        if callable(self._scaling):
            self._scaling(self)

    def load_parameters_from_database(self, use_default_removal=False):
        """
        Method to load parameters for from database.

        Args:
            use_default_removal - (optional) indicate whether to use defined
                                  default removal fraction if no specific value
                                  defined in database

        Returns:
            None
        """
        # Get parameter dict from database
        if self._tech_type is None:
            raise NotImplementedError(
                f"{self.name} derived zero order unit model has not "
                f"implemented the _tech_type attribute. This is required "
                f"to identify the database file to load parameters from.")

        # Get parameter dict from database
        pdict = self.config.database.get_unit_operation_parameters(
            self._tech_type, subtype=self.config.process_subtype)

        if self._has_recovery_removal:
            self.set_recovery_and_removal(pdict, use_default_removal)

        for v in self._fixed_perf_vars:
            self.set_param_from_data(v, pdict)

    def set_recovery_and_removal(self, data, use_default_removal=False):
        """
        Common utility method for setting values of recovery and removal
        fractions.

        Args:
            data - dict of parameter values to use when fixing variables
            use_default_removal - (optional) indicate whether to use defined
                                  default removal fraction if no specific value
                                  defined in database

        Returns:
            None
        """
        try:
            self.set_param_from_data(self.recovery_frac_mass_H2O, data)
        except KeyError:
            if self.recovery_frac_mass_H2O[:].fixed:
                pass
            else:
                raise

        for t, j in self.removal_frac_mass_solute:
            self.set_param_from_data(
                self.removal_frac_mass_solute[t, j],
                data,
                index=j,
                use_default_removal=use_default_removal)

    def set_param_from_data(
            self, parameter, data, index=None, use_default_removal=False):
        """
        General method for setting parameter values from a dict of data
        returned from a database.

        Args:
            parameter - a Pyomo Var to be fixed to value from database
            data - dict of parameter values from database
            index - (optional) index to fix if parameter is an IndexedVar
            use_default_removal - (optional) indicate whether to use defined
                                  default removal fraction if no specific value
                                  defined in database

        Returns:
            None

        Raises:
            KeyError if values cannot be found for parameter in data dict

        """

        pname = parameter.parent_component().local_name

        try:
            pdata = data[pname]
        except KeyError:
            raise KeyError(
                f"{self.name} - database provided does not contain an entry "
                f"for {pname} for technology.")

        if index is not None:
            try:
                pdata = pdata[index]
            except KeyError:
                if pname == "removal_frac_mass_solute" and use_default_removal:
                    try:
                        pdata = data["default_removal_frac_mass_solute"]
                        index = "default"
                    except KeyError:
                        raise KeyError(
                            f"{self.name} - database provided does not "
                            f"contain an entry for {pname} with index {index} "
                            f"for technology and no default removal was "
                            f"specified.")
                else:
                    raise KeyError(
                        f"{self.name} - database provided does not contain "
                        f"an entry for {pname} with index {index} for "
                        f"technology.")

        try:
            val = pdata["value"]
        except KeyError:
            raise KeyError(
                f"{self.name} - no value provided for {pname} (index: "
                f"{index}) in database.")
        try:
            units = getattr(pyunits, pdata["units"])
        except KeyError:
            raise KeyError(
                f"{self.name} - no units provided for {pname} (index: "
                f"{index}) in database.")

        parameter.fix(val*units)
        _log.info_high(f"{parameter.name} fixed to value {val} {str(units)}")

    def get_inlet_flow(self, t):
        return self._get_Q(self, t)

    def _get_stream_table_contents(self, time_point=0):
        return create_stream_table_dataframe(
            self._stream_table_dict,
            time_point=time_point)

    def _get_performance_contents(self, time_point=0):
        var_dict = {}

        for k, v in self._perf_var_dict.items():
            if k in ["Solute Removal", "Reaction Extent"]:
                for j, vd in v[time_point, :].wildcard_items():
                    var_dict[f"{k} [{j}]"] = vd
            elif v.is_indexed():
                var_dict[k] = v[time_point]
            else:
                var_dict[k] = v

        return {"vars": var_dict}



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
import numpy as np
import pyomo.environ as pyo
import sys
import os
import itertools
import warnings
import copy, pprint
import h5py

from scipy.interpolate import griddata
from enum import Enum, auto
from abc import abstractmethod, ABC
from idaes.core.util import get_solver

from idaes.surrogate.pysmo import sampling
from pyomo.common.collections import ComponentSet
from pyomo.common.tee import capture_output

np.set_printoptions(linewidth=200)

# ================================================================

class SamplingType(Enum):
    FIXED = auto()
    RANDOM = auto()
    RANDOM_LHS = auto()

# ================================================================

class _Sample(ABC):

    def __init__(self, pyomo_object, *args, **kwargs):
        # Check for indexed with single value
        if pyomo_object.is_indexed() and len(pyomo_object) == 1:
            for _data_obj in pyomo_object.values():
                pyomo_object = _data_obj

        # Make sure we are a Var() or Param()
        if not (pyomo_object.is_parameter_type() or pyomo_object.is_variable_type()):
            raise ValueError(f"The sweep parameter needs to be a pyomo Param or Var but {type(pyomo_object)} was provided instead.")
        if pyomo_object.is_parameter_type() and not pyomo_object.mutable:
            raise ValueError(f"Parameter {pyomo_object} is not mutable, and so cannot be set by parameter_sweep")
        self.pyomo_object = pyomo_object
        self.setup(*args, **kwargs)

    @abstractmethod
    def sample(self, num_samples):
        pass

    @abstractmethod
    def setup(self, *args, **kwargs):
        pass

# ================================================================

class RandomSample(_Sample):
    sampling_type = SamplingType.RANDOM

class FixedSample(_Sample):
    sampling_type = SamplingType.FIXED

# ================================================================

class LinearSample(FixedSample):

    def sample(self, num_samples):
        return np.linspace(self.lower_limit, self.upper_limit, self.num_samples)

    def setup(self, lower_limit, upper_limit, num_samples):
        self.lower_limit = lower_limit
        self.upper_limit = upper_limit
        self.num_samples = num_samples

# ================================================================

class UniformSample(RandomSample):

    def sample(self, num_samples):
        return np.random.uniform(self.lower_limit, self.upper_limit, num_samples)

    def setup(self, lower_limit, upper_limit):
        self.lower_limit = lower_limit
        self.upper_limit = upper_limit

# ================================================================

class NormalSample(RandomSample):

    def sample(self, num_samples):
        return np.random.normal(self.mean, self.sd, num_samples)

    def setup(self, mean, sd):
        self.mean = mean
        self.sd = sd

# ================================================================

class LatinHypercubeSample(_Sample):
    sampling_type = SamplingType.RANDOM_LHS

    def sample(self, num_samples):
        return [self.lower_limit, self.upper_limit]

    def setup(self, lower_limit, upper_limit):
        self.lower_limit = lower_limit
        self.upper_limit = upper_limit

# ================================================================

def _init_mpi(mpi_comm=None):

    if mpi_comm is None:
        try:
            from mpi4py import MPI
        except:
            warnings.warn("Could not import mpi4py from current environment (defaulting to serial).")
            return None, 0, 1

        else:
            mpi_comm = MPI.COMM_WORLD

    return mpi_comm, mpi_comm.Get_rank(), mpi_comm.Get_size()

# ================================================================

def _strip_extension(file_name, extension):
    if file_name.lower().endswith(extension):
        return file_name[:-len(extension)]
    else:
        return file_name

# ================================================================

def _build_combinations(d, sampling_type, num_samples, comm, rank, num_procs):
    num_var_params = len(d)

    if rank == 0:
        param_values = []

        for k, v in d.items():
            # Build a vector of discrete values for this parameter
            p = v.sample(num_samples)
            param_values.append(p)

        if sampling_type == SamplingType.FIXED:
            # Form an array with every possible combination of parameter values
            global_combo_array = np.array(np.meshgrid(*param_values, indexing="ij"))
            global_combo_array = global_combo_array.reshape(num_var_params, -1).T

        elif sampling_type == SamplingType.RANDOM:
            sorting = np.argsort(param_values[0])
            global_combo_array = np.vstack(param_values).T
            global_combo_array = global_combo_array[sorting, :]

        elif sampling_type == SamplingType.RANDOM_LHS:
            lb = [val[0] for val in param_values]
            ub = [val[1] for val in param_values]
            lhs = sampling.LatinHypercubeSampling([lb, ub], number_of_samples=num_samples, sampling_type='creation')
            global_combo_array = lhs.sample_points()
            sorting = np.argsort(global_combo_array[:, 0])
            global_combo_array = global_combo_array[sorting, :]

        else:
            raise ValueError(f"Unknown sampling type: {sampling_type}")

        # Test if the global_combo_array is in row-major order
        if not global_combo_array.flags.c_contiguous:
            # If not, return a copy of this array with row-major memory order
            global_combo_array = np.ascontiguousarray(global_combo_array)

    else:
        if sampling_type == SamplingType.FIXED:
            nx = 1
            for k, v in d.items():
                nx *= v.num_samples
        elif sampling_type == SamplingType.RANDOM or sampling_type == SamplingType.RANDOM_LHS:
            nx = num_samples
        else:
            raise ValueError(f"Unknown sampling type: {sampling_type}")

        if not float(nx).is_integer():
            raise RuntimeError(f"Total number of samples must be integer valued")
        nx = int(nx)

        # Allocate memory to hold the Bcast array
        global_combo_array = np.zeros((nx, num_var_params), dtype=np.float64)

    ### Broadcast the array to all processes
    if num_procs > 1:
        comm.Bcast(global_combo_array, root=0)

    return global_combo_array

# ================================================================

def _divide_combinations(global_combo_array, rank, num_procs):

    # Split the total list of combinations into NUM_PROCS chunks,
    # one per each of the MPI ranks
    # divided_combo_array = np.array_split(global_combo_array, num_procs, axis=0)
    divided_combo_array = np.array_split(global_combo_array, num_procs)

    # Return only this rank's portion of the total workload
    local_combo_array = divided_combo_array[rank]

    return local_combo_array

# ================================================================

def _update_model_values(m, param_dict, values):

    for k, item in enumerate(param_dict.values()):

        param = item.pyomo_object

        if param.is_variable_type():
            # Fix the single value to values[k]
            param.fix(values[k])

        elif param.is_parameter_type():
            # Fix the single value to values[k]
            param.set_value(values[k])

        else:
            raise RuntimeError(f"Unrecognized Pyomo object {param}")

# ================================================================

def _aggregate_results(local_results, global_values, comm, num_procs):

    if num_procs > 1: # pragma: no cover
        local_results = local_results.astype(np.float64)

        global_results = np.zeros((np.shape(global_values)[0], np.shape(local_results)[1]), dtype=np.float64)

        # Collect the number of result values to be sent from each process
        send_counts = np.zeros(num_procs, dtype=np.int64)
        comm.Gather(np.int64(np.size(local_results)), send_counts, root=0)

        # Collect the global results results onto rank 0
        comm.Gatherv(local_results, (global_results, send_counts), root=0)

        # Broadcast the results to all ranks
        comm.Bcast(global_results, root=0)

    else:
        global_results = np.copy(local_results)

    return global_results

# ================================================================

def _default_optimize(model, options=None, tee=False):
    '''
    Default optimization function used in parameter_sweep.
    Optimizes ``model`` using the IDAES default solver.
    Raises a RuntimeError if the TerminationCondition is not optimal

    Arguments:

        model : A Pyomo ConcreteModel to optimize

        options (optional) : Solver options to pass into idaes.core.utils.get_solver.
                             Default is None
        tee (options) : To display the solver log. Default it False

    '''
    solver = get_solver(options=options)
    results = solver.solve(model, tee=tee)
    return results

# ================================================================

def _process_sweep_params(sweep_params):

    sampling_type = None

    # Check the list of parameters to make sure they are valid
    for k in sweep_params:

        # Convert to using Sample class
        if isinstance(sweep_params[k], (list, tuple)):
            sweep_params[k] = LinearSample(*sweep_params[k])

        # Get the type of sampling
        current_sampling_type = sweep_params[k].sampling_type

        # Check to make sure only one sampling type is provided
        if sampling_type is None:
            sampling_type = current_sampling_type
        elif current_sampling_type != sampling_type:
            raise ValueError("Cannot mix sampling types")

    return sweep_params, sampling_type

# ================================================================

def _interp_nan_values(global_values, global_results):

    global_results_clean = np.copy(global_results)

    n_vals = np.shape(global_values)[1]
    n_outs = np.shape(global_results)[1]

    # Build a mask of all the non-nan saved outputs
    # i.e., where the optimzation succeeded
    mask = np.isfinite(global_results[:, 0])

    # Create a list of points where good data is available
    x0 = global_values[mask, :]

    if np.sum(mask) >= 4:
        # Interpolate to get a value for nan points where possible
        for k in range(n_outs):
            y0 = global_results[mask, k]
            yi = griddata(x0, y0, global_values, method='linear', rescale=True).reshape(-1)
            global_results_clean[~mask, k] = yi[~mask]

    else:
        warnings.warn("Too few points to perform interpolation.")


    return global_results_clean

# ================================================================

def _create_local_output_skeleton(model, sweep_params, outputs, num_samples):

    output_dict = {}
    output_dict["sweep_params"] = {}
    output_dict["outputs"] = {}

    sweep_param_objs = ComponentSet()

    # Store the inputs
    for sweep_param in sweep_params.values():
        var = sweep_param.pyomo_object
        sweep_param_objs.add(var)
        output_dict["sweep_params"][var.name] =  _create_component_output_skeleton(var, num_samples)

    if outputs is None:
        outputs = {}

        # No outputs are specified, so every Var, Expression, and Objective on the model should be saved
        for pyo_obj in model.component_data_objects((pyo.Var, pyo.Expression, pyo.Objective), active=True):
            # Only need to save this variable if it isn't one of the value in sweep_params
            if pyo_obj not in sweep_param_objs:
                output_dict["outputs"][pyo_obj.name] = _create_component_output_skeleton(pyo_obj, num_samples)
                outputs[pyo_obj.name] = pyo_obj

    else:
        # Save only the outputs specified in the outputs dictionary
        for short_name, pyo_obj in outputs.items():
            output_dict["outputs"][short_name] = _create_component_output_skeleton(pyo_obj, num_samples)

    return output_dict, outputs

# ================================================================

def _create_component_output_skeleton(component, num_samples):

    comp_dict = {}
    comp_dict["value"] = np.zeros(num_samples, dtype=np.float)
    if hasattr(component, 'lb'):
        comp_dict["lower bound"] = component.lb
    if hasattr(component, 'ub'):
        comp_dict["upper bound"] = component.lb
    if hasattr(component, 'get_units'):
        unit_obj = component.get_units()
        if unit_obj is not None:
            comp_dict["units"] = component.get_units().name
        else:
            comp_dict["units"] = "None"

    return comp_dict

# ================================================================

def _update_local_output_dict(model, sweep_params, case_number, sweep_vals, run_successful, output_dict, outputs):

    # Get the inputs
    op_ps_dict = output_dict["sweep_params"]
    for key, item in sweep_params.items():
        var_name = item.pyomo_object.name
        op_ps_dict[var_name]['value'][case_number] = item.pyomo_object.value

    # Get the outputs from model
    if run_successful:
        for label, pyo_obj in outputs.items():
            output_dict["outputs"][label]["value"][case_number] = pyo.value(pyo_obj)

    else:
        for label in outputs.keys():
            output_dict["outputs"][label]["value"][case_number] = np.nan

# ================================================================

def _create_global_output(local_output_dict, req_num_samples, comm, rank, num_procs):

    if num_procs == 1:
        global_output_dict = local_output_dict
    else: # pragma: no cover
        # We make the assumption that the parameter sweep is running the same
        # flowsheet num_samples number of times, i.e., the structure of the
        # local_output_dict remains the same across all mpi_ranks
        local_num_cases = len(local_output_dict["solve_successful"])

        # Gather the size of the value array on each MPI rank
        sample_split_arr = comm.allgather(local_num_cases)
        num_total_samples = sum(sample_split_arr)

        # Create the global value array on rank 0
        if rank == 0:
            global_output_dict = copy.deepcopy(local_output_dict)
            # Create a global value array of inputs in the dictionary
            for key, item in global_output_dict.items():
                if key != "solve_successful":
                    for subkey, subitem in item.items():
                        subitem['value'] = np.zeros(num_total_samples, dtype=np.float)

        else:
            global_output_dict = local_output_dict

        # Finally collect the values
        for key, item in local_output_dict.items(): # This probably doesnt work
            if key != "solve_successful":
                for subkey, subitem in item.items():
                    comm.Gatherv(sendbuf=subitem["value"],
                                 recvbuf=(global_output_dict[key][subkey]["value"], sample_split_arr),
                                 root=0)

                    # Trim to the exact number
                    global_output_dict[key][subkey]["value"] = global_output_dict[key][subkey]["value"][0:req_num_samples]

            elif key == "solve_successful":
                local_solve_successful = np.fromiter(item, dtype=np.bool, count=len(item))

                if rank == 0:
                    global_solve_successful = np.empty(num_total_samples, dtype=np.bool)
                else:
                    global_solve_successful = None

                comm.Gatherv(sendbuf=local_solve_successful,
                             recvbuf=(global_solve_successful, sample_split_arr),
                             root=0)

                if rank == 0:
                    global_output_dict[key] = global_solve_successful[0:req_num_samples]

    return global_output_dict

# ================================================================

def _write_outputs(output_dict, output_directory, h5_results_file, txt_options="metadata"):

    if not h5_results_file.endswith(".h5"):
        h5_results_file += ".h5"

    _write_output_to_h5(output_dict, output_directory, h5_results_file)

    # We will also create a companion txt file by default which contains
    # the metadata of the h5 file in a user readable format.
    txt_fname = _strip_extension(h5_results_file,".h5") + ".txt"
    txt_fpath = os.path.join(output_directory, txt_fname)
    if "solve_successful" in output_dict.keys():
        output_dict.pop("solve_successful")
    if txt_options == "metadata":
        my_dict = copy.deepcopy(output_dict)
        for key, value in my_dict.items():
            for subkey, subvalue in value.items():
                subvalue.pop('value')
    elif txt_options == "keys":
        my_dict = {}
        for key, value in output_dict.items():
            my_dict[key] = list(value.keys())
    else:
        my_dict = output_dict

    with open(txt_fpath, "w") as log_file:
        pprint.pprint(my_dict, log_file)

# ================================================================

def _write_output_to_h5(output_dict, output_directory, fname):

    fpath = os.path.join(output_directory, fname)
    f = h5py.File(fpath, 'w')
    for key, item in output_dict.items():
        grp = f.create_group(key)
        if key != "solve_successful":
            for subkey, subitem in item.items():
                subgrp = grp.create_group(subkey)
                for subsubkey, subsubitem in subitem.items():
                    if subsubkey == 'lower bound' and subsubitem is None:
                        subgrp.create_dataset(subsubkey, data=np.finfo('d').min)
                    elif subsubkey == 'upper bound' and subsubitem is None:
                        subgrp.create_dataset(subsubkey, data=np.finfo('d').max)
                    else:
                        subgrp.create_dataset(subsubkey, data=output_dict[key][subkey][subsubkey])
        elif key == 'solve_successful':
            grp.create_dataset(key, data=output_dict[key])

    f.close()

# ================================================================

def _read_output_h5(filepath):

    f = h5py.File(filepath , 'r')

    l1_keys = list(f.keys())
    output_dict = {}
    for key in l1_keys: # Input or Output
        if key != 'solve_successful':
            output_dict[key] = {}
            l2_keys = list(f[key].keys())
            for subkey in l2_keys: # Variable name
                output_dict[key][subkey] = {}
                l3_keys = list(f[key][subkey].keys())
                for subsubkey in l3_keys: # variable metadata
                    output_dict[key][subkey][subsubkey] = f[key][subkey][subsubkey][()]
                    if subsubkey == "units":
                        # The strings are recovered in bytes. we choose to convert it to utf-8
                        output_dict[key][subkey][subsubkey] = output_dict[key][subkey][subsubkey].decode("utf-8")
        elif key == 'solve_successful':
            output_dict[key] = list(f[key]['solve_successful'][()])

    f.close()

    return output_dict

# ================================================================

def _do_param_sweep(model, sweep_params, outputs, local_values, optimize_function, optimize_kwargs,
        reinitialize_function, reinitialize_kwargs, reinitialize_before_sweep, comm):

    # Initialize space to hold results
    local_num_cases = np.shape(local_values)[0]

    # Create the output skeleton for storing detailed data
    local_output_dict, outputs = _create_local_output_skeleton(model, sweep_params, outputs, local_num_cases)

    local_results = np.zeros((local_num_cases, len(outputs)))

    local_solve_successful_list = []

    # ================================================================
    # Run all optimization cases
    # ================================================================

    for k in range(local_num_cases):
        # Update the model values with a single combination from the parameter space
        _update_model_values(model, sweep_params, local_values[k, :])

        run_successful = False #until proven otherwise

        # Forced reinitialization of the flowsheet if enabled
        if reinitialize_before_sweep:
            try:
                assert reinitialize_function is not None
            except:
                raise ValueError("Reinitialization function was not specified. The model will not be reinitialized.")
            else:
                reinitialize_function(model, **reinitialize_kwargs)

        try:
            # Simulate/optimize with this set of parameter
            with capture_output():
                results = optimize_function(model, **optimize_kwargs)
            pyo.assert_optimal_termination(results)

        except:
            # If the run is infeasible, report nan
            local_results[k, :] = np.nan

        else:
            # If the simulation suceeds, report stats
            local_results[k, :] = [pyo.value(outcome) for outcome in outputs.values()]
            run_successful = True

        # If the initial attempt failed and additional conditions are met, try
        # to reinitialize and resolve.
        if not run_successful and (reinitialize_function is not None):
            try:
                reinitialize_function(model, **reinitialize_kwargs)
                with capture_output():
                    results = optimize_function(model, **optimize_kwargs)
                pyo.assert_optimal_termination(results)

            except:
                pass
            else:
                local_results[k, :] = [pyo.value(outcome) for outcome in outputs.values()]
                run_successful = True

        # Update the loop based on the reinitialization
        _update_local_output_dict(model, sweep_params, k, local_values[k, :], run_successful, local_output_dict, outputs)

        local_solve_successful_list.append(run_successful)

    local_output_dict["solve_successful"] = local_solve_successful_list

    return local_results, local_output_dict

# ================================================================

def _aggregate_local_results(global_values, local_results, local_output_dict,
        num_samples, local_num_cases, comm, rank, num_procs):

    global_results = _aggregate_results(local_results, global_values, comm, num_procs)
    global_output_dict = _create_global_output(local_output_dict, num_samples, comm, rank, num_procs)

    return global_results, global_output_dict

# ================================================================

def _save_results(sweep_params, outputs, local_values, global_values, local_results,
        global_results, global_output_dict, csv_results_file, h5_results_file,
        debugging_data_dir, comm, rank, num_procs, interpolate_nan_outputs):

    # Make a directory for saved outputs
    if rank == 0:
        if csv_results_file is not None:
            if not csv_results_file.endswith(".csv"):
                csv_results_file += ".csv"
            dirname = os.path.dirname(csv_results_file)
            if dirname != '':
                os.makedirs(dirname, exist_ok=True)

        if debugging_data_dir is not None:
            os.makedirs(debugging_data_dir, exist_ok=True)

    if num_procs > 1:
        comm.Barrier()

    # Write a header string for all data files
    data_header = ','.join(itertools.chain(sweep_params,global_output_dict['outputs']))

    if debugging_data_dir is not None:
        # Create the local filename and data
        fname = os.path.join(debugging_data_dir, f'local_results_{rank:03}.csv')
        local_save_data = np.hstack((local_values, local_results))

        # Save the local data
        np.savetxt(fname, local_save_data, header=data_header, delimiter=', ', fmt='%.6e')

    # Create the global filename and data
    global_save_data = np.hstack((global_values, global_results))

    if rank == 0 and csv_results_file is not None:
        # Save the global data
        np.savetxt(csv_results_file, global_save_data, header=data_header, delimiter=',', fmt='%.6e')

        if interpolate_nan_outputs:
            global_results_clean = _interp_nan_values(global_values, global_results)
            global_save_data_clean = np.hstack((global_values, global_results_clean))

            head, tail = os.path.split(csv_results_file)

            if head == '':
                interp_file = 'interpolated_%s' % (tail)
            else:
                interp_file = '%s/interpolated_%s' % (head, tail)

            np.savetxt(interp_file, global_save_data_clean, header=data_header, delimiter=',', fmt='%.6e')

    if rank == 0 and h5_results_file is not None:
        # Save the data of output dictionary
        _write_outputs(global_output_dict, dirname, h5_results_file, txt_options="keys")

    return global_save_data


# ================================================================

def parameter_sweep(model, sweep_params, outputs=None, csv_results_file=None, h5_results_file=None,
        optimize_function=_default_optimize, optimize_kwargs=None, reinitialize_function=None,
        reinitialize_kwargs=None, reinitialize_before_sweep=False, mpi_comm=None, debugging_data_dir=None,
        interpolate_nan_outputs=False, num_samples=None, seed=None):

    '''
    This function offers a general way to perform repeated optimizations
    of a model for the purposes of exploring a parameter space while
    monitoring multiple outputs.
    If provided, writes single CSV file to ``results_file`` with all inputs and resulting outputs.

    Arguments:

        model : A Pyomo ConcreteModel containing a watertap flowsheet, for best
                results it should be initialized before being passed to this
                function.

        sweep_params: A dictionary containing the values to vary with the format
                      ``sweep_params['Short/Pretty-print Name'] =
                      (model.fs.variable_or_param[index], lower_limit, upper_limit, num_samples)``.
                      A uniform number of samples ``num_samples`` will be take between
                      the ``lower_limit`` and ``upper_limit``.

        outputs : An optional dictionary containing "short names" as keys and and Pyomo objects
                  on ``model`` whose values to report as values. E.g.,
                  ``outputs['Short/Pretty-print Name'] = model.fs.variable_or_expression_to_report``.
                  If not provided, i.e., outputs = None, the default behavior is to save all model
                  variables, parameters, and expressions which provides very thorough results
                  at the cost of large file sizes.

        csv_results_file (optional) : The path and file name where the results are to be saved;
                                   subdirectories will be created as needed.

        h5_results_file (optional) : The file name without the extension where the results are to be saved;
                                   The path is identified from the arguments of `csv_results_file`. This
                                   filename is used when creating the H5 file and the companion text file
                                   which contains the variable names contained within the H5 file.

        optimize_function (optional) : A user-defined function to perform the optimization of flowsheet
                                       ``model`` and loads the results back into ``model``. The first
                                       argument of this function is ``model``\. The default uses the
                                       default IDAES solver, raising an exception if the termination
                                       condition is not optimal.

        optimize_kwargs (optional) : Dictionary of kwargs to pass into every call to
                                     ``optimize_function``. The first arg will always be ``model``,
                                     e.g., ``optimize_function(model, **optimize_kwargs)``. The default
                                     uses no kwargs.

        reinitialize_function (optional) : A user-defined function to perform the re-initialize the
                                           flowsheet ``model`` if the first call to ``optimize_function``
                                           fails for any reason. After ``reinitialize_function``, the
                                           parameter sweep tool will immediately call
                                           ``optimize_function`` again.

        reinitialize_kwargs (optional) : Dictionary or kwargs to pass into every call to
                                         ``reinitialize_function``. The first arg will always be
                                         ``model``, e.g.,
                                         ``reinitialize_function(model, **reinitialize_kwargs)``.
                                         The default uses no kwargs.

        reinitialize_before_sweep (optional): Boolean option to reinitialize the flow sheet model before
                                              every parameter sweep realization. The default is False.
                                              Note the parameter sweep model will try to reinitialize the
                                              solve regardless of the option if the run fails.

        mpi_comm (optional) : User-provided MPI communicator for parallel parameter sweeps.
                              If None COMM_WORLD will be used. The default is sufficient for most
                              users.

        debugging_data_dir (optional) : Save results on a per-process basis for parallel debugging
                                        purposes. If None no `debugging` data will be saved.

        interpolate_nan_outputs (optional) : When the parameter sweep has finished, interior values
                                             of np.nan will be replaced with a value obtained via
                                             a linear interpolation of their surrounding valid neighbors.
                                             If true, a second output file with the extension "_clean"
                                             will be saved alongside the raw (un-interpolated) values.

        num_samples (optional) : If the user is using sampling techniques rather than a linear grid
                                 of values, they need to set the number of samples

        seed (optional) : If the user is using a random sampling technique, this sets the seed

    Returns:

        save_data : A list were the first N columns are the values of the parameters passed
                    by ``sweep_params`` and the remaining columns are the values of the
                    simulation identified by the ``outputs`` argument.
    '''

    # Get an MPI communicator
    comm, rank, num_procs = _init_mpi(mpi_comm)

    # Convert sweep_params to LinearSamples
    sweep_params, sampling_type = _process_sweep_params(sweep_params)

    # Set the seed before sampling
    np.random.seed(seed)

    # Enumerate/Sample the parameter space
    global_values = _build_combinations(sweep_params, sampling_type, num_samples, comm, rank, num_procs)

    # divide the workload between processors
    local_values = _divide_combinations(global_values, rank, num_procs)
    local_num_cases = np.shape(local_values)[0]

    # Set up optimize_kwargs
    if optimize_kwargs is None:
        optimize_kwargs = dict()
    # Set up reinitialize_kwargs
    if reinitialize_kwargs is None:
        reinitialize_kwargs = dict()

    # Do the Loop
    local_results, local_output_dict = _do_param_sweep(model, sweep_params, outputs, local_values,
        optimize_function, optimize_kwargs, reinitialize_function, reinitialize_kwargs, reinitialize_before_sweep,
        comm)

    # Aggregate results on Master
    global_results, global_output_dict = _aggregate_local_results(global_values, local_results, local_output_dict,
            num_samples, local_num_cases, comm, rank, num_procs)

    # Save to file
    global_save_data = _save_results(sweep_params, outputs, local_values, global_values, local_results, global_results, global_output_dict,
        csv_results_file, h5_results_file, debugging_data_dir, comm, rank, num_procs, interpolate_nan_outputs)

    return global_save_data

# ================================================================



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
This module contains a zero-order representation of an ion exchange unit.
operation.
"""

from pyomo.environ import Reference, units as pyunits, Var
from idaes.core import declare_process_block_class
from watertap.core import build_sido, pump_electricity, ZeroOrderBaseData

# Some more information about this module
__author__ = "Adam Atia"


@declare_process_block_class("IonExchangeZO")
class IonExchangeZOData(ZeroOrderBaseData):
    """
    Zero-Order model for an Ion exchange unit operation.
    """

    CONFIG = ZeroOrderBaseData.CONFIG()

    def build(self):
        super().build()

        self._tech_type = "ion_exchange"

        build_sido(self)
        self._Q = Reference(self.properties_in[:].flow_vol)
        pump_electricity(self, self._Q)

        # mutable parameter; default value found in WT3 for anion exchange
        self.eta_pump.set_value(0.8)
        # mutable parameter; default value of 2 bar converted to feet head
        self.lift_height.set_value(69.91052 * pyunits.feet)

        # Add variables and constraints for material requirements
        self.NaCl_flowrate = Var(self.flowsheet().time,
                                 initialize=1,
                                 units=pyunits.kg/pyunits.s,
                                 bounds=(0, None),
                                 doc="Flowrate of NaCl addition")
        self.NaCl_dose = Var(units=pyunits.kg/pyunits.m**3,
                             bounds=(0, None),
                             doc="Dosage of NaCl addition")

        self._fixed_perf_vars.append(self.NaCl_dose)
        self._perf_var_dict["NaCl Addition"] = self.NaCl_flowrate

        @self.Constraint(self.flowsheet().time)
        def NaCl_constraint(blk, t):
            return (blk.NaCl_flowrate[t] ==
                    blk.NaCl_dose*blk.properties_in[t].flow_vol)

        self.resin_demand = Var(
            self.flowsheet().time,
            initialize=1,
            units=pyunits.kg/pyunits.s,
            bounds=(0, None),
            doc="Replacement rate of ion exchange resin")
        self.resin_replacement = Var(
            units=pyunits.kg/pyunits.m**3,
            bounds=(0, None),
            doc="Resin replacement as a function of flow")

        self._fixed_perf_vars.append(self.resin_replacement)
        self._perf_var_dict["Resin Demand"] = self.resin_demand

        @self.Constraint(self.flowsheet().time)
        def resin_constraint(blk, t):
            return (blk.resin_demand[t] ==
                    blk.resin_replacement*blk.properties_in[t].flow_vol)

        if self.config.process_subtype == "clinoptilolite":
            if "ammonium_as_nitrogen" in self.config.property_package.solute_set:
                self.nitrogen_clay_ratio = Var(self.flowsheet().config.time,
                                               units=pyunits.dimensionless,
                                               doc="Mass fraction of nitrogen in clay mixture")

                self._fixed_perf_vars.append(self.nitrogen_clay_ratio)

                self.final_solids_mass = Var(self.flowsheet().config.time,
                                             units=pyunits.kg / pyunits.s,
                                             doc="Solids mass flow in byproduct stream")

                @self.Constraint(self.flowsheet().time,
                                 doc="Solids mass flow in byproduct stream constraint")
                def solids_mass_flow_constraint(b, t):
                    return (b.final_solids_mass[t] ==
                            b.properties_byproduct[t].flow_mass_comp["ammonium_as_nitrogen"]
                            / b.nitrogen_clay_ratio[t])

                self._perf_var_dict["Nitrogen-Clay Mixture Ratio (kg/kg)"] = self.nitrogen_clay_ratio
                self._perf_var_dict["Final mass flow of clay and nitrogen (kg/s)"] = self.final_solids_mass
            else:
                raise KeyError("ammonium_as_nitrogen should be defined in solute_list for this subtype.")



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################

from copy import deepcopy
from enum import Enum, auto
from pyomo.environ import Block, exp, NonNegativeReals, Param, Suffix, Var, units as pyunits
from pyomo.common.collections import ComponentSet
from pyomo.common.config import ConfigBlock, ConfigValue, In
from idaes.core import UnitModelBlockData, useDefault, MaterialBalanceType,\
        EnergyBalanceType, MomentumBalanceType
from idaes.core.util import scaling as iscale
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.exceptions import ConfigurationError
from idaes.core.util.tables import create_stream_table_dataframe
import idaes.logger as idaeslog

_log = idaeslog.getLogger(__name__)


class ConcentrationPolarizationType(Enum):
    none = auto()                    # simplified assumption: no concentration polarization
    fixed = auto()                   # simplified assumption: concentration polarization modulus is a user specified value
    calculated = auto()              # calculate concentration polarization (concentration at membrane interface)


class MassTransferCoefficient(Enum):
    none = auto()                    # mass transfer coefficient not utilized for concentration polarization effect
    fixed = auto()                   # mass transfer coefficient is a user specified value
    calculated = auto()              # mass transfer coefficient is calculated
    # TODO: add option for users to define their own relationship?


class PressureChangeType(Enum):
    fixed_per_stage = auto()         # pressure drop across membrane channel is a user-specified value
    fixed_per_unit_length = auto()   # pressure drop per unit length across membrane channel is a user-specified value
    calculated = auto()              # pressure drop across membrane channel is calculated


class _ReverseOsmosisBaseData(UnitModelBlockData):
    """
    Reverse Osmosis base class
    """

    CONFIG = ConfigBlock()

    CONFIG.declare("dynamic", ConfigValue(
        default=False,
        domain=In([False]),
        description="Dynamic model flag - must be False",
        doc="""Indicates whether this model will be dynamic or not.
    **default** = False. Membrane units do not yet support dynamic
    behavior."""))

    CONFIG.declare("has_holdup", ConfigValue(
            default=False,
            domain=In([False]),
            description="Holdup construction flag - must be False",
            doc="""Indicates whether holdup terms should be constructed or not.
    **default** - False. Membrane units do not have defined volume, thus
    this must be False."""))

    CONFIG.declare("property_package", ConfigValue(
        default=useDefault,
        domain=is_physical_parameter_block,
        description="Property package to use for control volume",
        doc="""Property parameter object used to define property calculations,
    **default** - useDefault.
    **Valid values:** {
    **useDefault** - use default package from parent model or flowsheet,
    **PhysicalParameterObject** - a PhysicalParameterBlock object.}"""))

    CONFIG.declare("property_package_args", ConfigBlock(
        implicit=True,
        description="Arguments to use for constructing property packages",
        doc="""A ConfigBlock with arguments to be passed to a property block(s)
    and used when constructing these.
    **default** - None.
    **Valid values:** {
    see property package for documentation.}"""))

    CONFIG.declare("material_balance_type", ConfigValue(
            default=MaterialBalanceType.useDefault,
            domain=In(MaterialBalanceType),
            description="Material balance construction flag",
            doc="""Indicates what type of mass balance should be constructed,
    **default** - MaterialBalanceType.useDefault.
    **Valid values:** {
    **MaterialBalanceType.useDefault - refer to property package for default
    balance type
    **MaterialBalanceType.none** - exclude material balances,
    **MaterialBalanceType.componentPhase** - use phase component balances,
    **MaterialBalanceType.componentTotal** - use total component balances,
    **MaterialBalanceType.elementTotal** - use total element balances,
    **MaterialBalanceType.total** - use total material balance.}"""))

    CONFIG.declare("energy_balance_type", ConfigValue(
        default=EnergyBalanceType.useDefault,
        domain=In(EnergyBalanceType),
        description="Energy balance construction flag",
        doc="""Indicates what type of energy balance should be constructed.
    **default** - EnergyBalanceType.useDefault.
    **Valid values:** {
    **EnergyBalanceType.useDefault - refer to property package for default
    balance type
    **EnergyBalanceType.none** - exclude energy balances,
    **EnergyBalanceType.enthalpyTotal** - single enthalpy balance for material,
    **EnergyBalanceType.enthalpyPhase** - enthalpy balances for each phase,
    **EnergyBalanceType.energyTotal** - single energy balance for material,
    **EnergyBalanceType.energyPhase** - energy balances for each phase.}"""))

    CONFIG.declare("momentum_balance_type", ConfigValue(
            default=MomentumBalanceType.pressureTotal,
            domain=In(MomentumBalanceType),
            description="Momentum balance construction flag",
            doc="""Indicates what type of momentum balance should be constructed,
    **default** - MomentumBalanceType.pressureTotal.
    **Valid values:** {
    **MomentumBalanceType.none** - exclude momentum balances,
    **MomentumBalanceType.pressureTotal** - single pressure balance for material,
    **MomentumBalanceType.pressurePhase** - pressure balances for each phase,
    **MomentumBalanceType.momentumTotal** - single momentum balance for material,
    **MomentumBalanceType.momentumPhase** - momentum balances for each phase.}"""))

    CONFIG.declare("concentration_polarization_type", ConfigValue(
        default=ConcentrationPolarizationType.calculated,
        domain=In(ConcentrationPolarizationType),
        description="External concentration polarization effect in RO",
        doc="""
            Options to account for concentration polarization.

            **default** - ``ConcentrationPolarizationType.calculated``

        .. csv-table::
            :header: "Configuration Options", "Description"

            "``ConcentrationPolarizationType.none``", "Simplifying assumption to ignore concentration polarization"
            "``ConcentrationPolarizationType.fixed``", "Specify an estimated value for the concentration polarization modulus"
            "``ConcentrationPolarizationType.calculated``", "Allow model to perform calculation of membrane-interface concentration"
        """))

    CONFIG.declare("mass_transfer_coefficient", ConfigValue(
        default=MassTransferCoefficient.calculated,
        domain=In(MassTransferCoefficient),
        description="Mass transfer coefficient in RO feed channel",
        doc="""
            Options to account for mass transfer coefficient.

            **default** - ``MassTransferCoefficient.calculated``

        .. csv-table::
            :header: "Configuration Options", "Description"

            "``MassTransferCoefficient.none``", "Mass transfer coefficient not used in calculations"
            "``MassTransferCoefficient.fixed``", "Specify an estimated value for the mass transfer coefficient in the feed channel"
            "``MassTransferCoefficient.calculated``", "Allow model to perform calculation of mass transfer coefficient"
        """))

    CONFIG.declare("has_pressure_change", ConfigValue(
            default=False,
            domain=In([True, False]),
            description="Pressure change term construction flag",
            doc="""Indicates whether terms for pressure change should be
    constructed,
    **default** - False.
    **Valid values:** {
    **True** - include pressure change terms,
    **False** - exclude pressure change terms.}"""))

    CONFIG.declare("pressure_change_type", ConfigValue(
        default=PressureChangeType.fixed_per_stage,
        domain=In(PressureChangeType),
        description="Pressure change term construction flag",
        doc="""
        Indicates what type of pressure change calculation will be made. To use any of the 
        ``pressure_change_type`` options to account for pressure drop, the configuration keyword 
        ``has_pressure_change`` must also be set to ``True``. Also, if a value is specified for pressure 
        change, it should be negative to represent pressure drop. 
        
        **default** - ``PressureChangeType.fixed_per_stage`` 


    .. csv-table::
        :header: "Configuration Options", "Description"

        "``PressureChangeType.fixed_per_stage``", "Specify an estimated value for pressure drop across the membrane feed channel"
        "``PressureChangeType.fixed_per_unit_length``", "Specify an estimated value for pressure drop per unit length across the membrane feed channel"
        "``PressureChangeType.calculated``", "Allow model to perform calculation of pressure drop across the membrane feed channel"

"""))

    CONFIG.declare("has_full_reporting", ConfigValue(
            default=False,
            domain=In([True, False]),
            description="Level of reporting results",
            doc="""Level of reporting results.
            **default** - False.
            **Valid values:** {
            **False** - include minimal reporting of results,
            **True** - report additional properties of interest that aren't constructed by
            the unit model by default. Also, report averaged expression values"""))


    def build(self):
        """
        Common variables and constraints for an RO unit model
        """
        super().build()

        if len(self.config.property_package.solvent_set) > 1:
            raise ConfigurationError("Membrane models only support one solvent component,"
                                     "the provided property package has specified {} solvent components"
                                     .format(len(self.config.property_package.solvent_set)))

        if (len(self.config.property_package.phase_list) > 1
                or 'Liq' not in [p for p in self.config.property_package.phase_list]):
            raise ConfigurationError(
                "Membrane models only support one liquid phase ['Liq'],"
                "the property package has specified the following phases {}"
                .format([p for p in self.config.property_package.phase_list]))

        if (self.config.pressure_change_type is not PressureChangeType.fixed_per_stage
                and self.config.has_pressure_change is False):
            raise ConfigurationError(
                "\nConflict between configuration options:\n"
                "'has_pressure_change' cannot be False "
                "while 'pressure_change_type' is set to {}.\n\n"
                "'pressure_change_type' must be set to PressureChangeType.fixed_per_stage\nor "
                "'has_pressure_change' must be set to True"
                .format(self.config.pressure_change_type))

        if (self.config.concentration_polarization_type == ConcentrationPolarizationType.calculated
                and self.config.mass_transfer_coefficient == MassTransferCoefficient.none):
            raise ConfigurationError(
                "\n'mass_transfer_coefficient' and 'concentration_polarization_type' options configured incorrectly:\n"
                "'mass_transfer_coefficient' cannot be set to MassTransferCoefficient.none "
                "while 'concentration_polarization_type' is set to ConcentrationPolarizationType.calculated.\n "
                "\n\nSet 'mass_transfer_coefficient' to MassTransferCoefficient.fixed or "
                "MassTransferCoefficient.calculated "
                "\nor set 'concentration_polarization_type' to ConcentrationPolarizationType.fixed or "
                "ConcentrationPolarizationType.none")

        if (self.config.concentration_polarization_type != ConcentrationPolarizationType.calculated
                and self.config.mass_transfer_coefficient != MassTransferCoefficient.none):
            raise ConfigurationError(
                "\nConflict between configuration options:\n"
                "'mass_transfer_coefficient' cannot be set to {} "
                "while 'concentration_polarization_type' is set to {}.\n\n"
                "'mass_transfer_coefficient' must be set to MassTransferCoefficient.none\nor "
                "'concentration_polarization_type' must be set to ConcentrationPolarizationType.calculated"
                    .format(self.config.mass_transfer_coefficient, self.config.concentration_polarization_type))

        self.scaling_factor = Suffix(direction=Suffix.EXPORT)

    def _make_performance(self):

        # For permeate-specific scaling in calculate_scaling_factors
        self._permeate_scaled_properties = ComponentSet()

        solvent_set = self.config.property_package.solvent_set
        solute_set = self.config.property_package.solute_set

        units_meta = \
            self.config.property_package.get_metadata().get_derived_units

        self.nfe = Param(
            initialize=(len(self.difference_elements)),
            units=pyunits.dimensionless,
            doc="Number of finite elements")

        """ Unit model variables"""
        self.A_comp = Var(
            self.flowsheet().config.time,
            solvent_set,
            initialize=1e-12,
            bounds=(1e-18, 1e-6),
            domain=NonNegativeReals,
            units=units_meta('length') * units_meta('pressure') ** -1 * units_meta('time') ** -1,
            doc='Solvent permeability coeff.')

        self.B_comp = Var(
            self.flowsheet().config.time,
            solute_set,
            initialize=1e-8,
            bounds=(1e-11, 1e-5),
            domain=NonNegativeReals,
            units=units_meta('length')*units_meta('time')**-1,
            doc='Solute permeability coeff.')

        # TODO: add water density to NaCl prop model and remove here (or use IDAES version)
        self.dens_solvent = Param(
            initialize=1000,
            units=units_meta('mass')*units_meta('length')**-3,
            doc='Pure water density')

        self.area = Var(
            initialize=10,
            bounds=(1e-1, 1e3),
            domain=NonNegativeReals,
            units=units_meta('length')**2,
            doc='Membrane area')

        self.recovery_mass_phase_comp = Var(
            self.flowsheet().config.time,
            self.config.property_package.phase_list,
            self.config.property_package.component_list,
            initialize=lambda b,t,p,j : 0.4037 if j in solvent_set else 0.0033,
            bounds=lambda b,t,p,j : (1e-2, 1 - 1e-6) if j in solvent_set else (1e-5, 1 - 1e-6),
            units=pyunits.dimensionless,
            doc='Mass-based component recovery')

        self.recovery_vol_phase = Var(
            self.flowsheet().config.time,
            self.config.property_package.phase_list,
            initialize=0.4,
            bounds=(1e-2, 1 - 1e-6),
            units=pyunits.dimensionless,
            doc='Volumetric recovery rate')

        self.rejection_phase_comp = Var(
            self.flowsheet().config.time,
            self.config.property_package.phase_list,
            solute_set,
            initialize=0.9,
            bounds=(1e-2, 1 - 1e-6),
            units=pyunits.dimensionless,
            doc='Observed solute rejection')

        self.flux_mass_phase_comp = Var(
            self.flowsheet().config.time,
            self.length_domain,
            self.config.property_package.phase_list,
            self.config.property_package.component_list,
            initialize=lambda b,t,x,p,j : 5e-4 if j in solvent_set else 1e-6,
            bounds=lambda b,t,x,p,j : (1e-4, 3e-2) if j in solvent_set else (1e-8, 1e-3),
            units=units_meta('mass')*units_meta('length')**-2*units_meta('time')**-1,
            doc='Mass flux across membrane at inlet and outlet')

        if ((self.config.mass_transfer_coefficient == MassTransferCoefficient.calculated)
                or self.config.pressure_change_type == PressureChangeType.calculated):

            self.channel_height = Var(
                initialize=1e-3,
                bounds=(1e-4, 5e-3),
                domain=NonNegativeReals,
                units=units_meta('length'),
                doc='Feed-channel height')

            self.dh = Var(
                initialize=1e-3,
                bounds=(1e-4, 5e-3),
                domain=NonNegativeReals,
                units=units_meta('length'),
                doc='Hydraulic diameter of feed channel')

            self.spacer_porosity = Var(
                initialize=0.95,
                bounds=(0.1, 0.99),
                domain=NonNegativeReals,
                units=pyunits.dimensionless,
                doc='Feed-channel spacer porosity')

        if self.config.concentration_polarization_type == ConcentrationPolarizationType.calculated:
            self.Kf = Var(
                self.flowsheet().config.time,
                self.length_domain,
                solute_set,
                initialize=5e-5,
                bounds=(1e-6, 1e-3),
                domain=NonNegativeReals,
                units=units_meta('length') * units_meta('time')**-1,
                doc='Mass transfer coefficient in feed channel')
        if (self.config.mass_transfer_coefficient == MassTransferCoefficient.calculated
                or self.config.pressure_change_type == PressureChangeType.calculated):
            self.N_Re = Var(
                self.flowsheet().config.time,
                self.length_domain,
                initialize=5e2,
                bounds=(10, 5e3),
                domain=NonNegativeReals,
                units=pyunits.dimensionless,
                doc="Reynolds number in feed channel")
        if self.config.mass_transfer_coefficient == MassTransferCoefficient.calculated:
            self.N_Sc = Var(
                self.flowsheet().config.time,
                self.length_domain,
                initialize=5e2,
                bounds=(1e2, 2e3),
                domain=NonNegativeReals,
                units=pyunits.dimensionless,
                doc="Schmidt number in feed channel")
            self.N_Sh = Var(
                self.flowsheet().config.time,
                self.length_domain,
                initialize=1e2,
                bounds=(1, 3e2),
                domain=NonNegativeReals,
                units=pyunits.dimensionless,
                doc="Sherwood number in feed channel")

        if self.config.pressure_change_type == PressureChangeType.calculated:
            self.velocity = Var(
                self.flowsheet().config.time,
                self.length_domain,
                initialize=0.5,
                bounds=(1e-2, 5),
                domain=NonNegativeReals,
                units=units_meta('length')/units_meta('time'),
                doc="Crossflow velocity in feed channel")
            self.friction_factor_darcy = Var(
                self.flowsheet().config.time,
                self.length_domain,
                initialize=0.5,
                bounds=(1e-2, 5),
                domain=NonNegativeReals,
                units=pyunits.dimensionless,
                doc="Darcy friction factor in feed channel")

        if (self.config.mass_transfer_coefficient == MassTransferCoefficient.calculated
            or self.config.pressure_change_type == PressureChangeType.calculated):

            @self.Constraint(doc="Hydraulic diameter")  # eqn. 17 in Schock & Miquel, 1987
            def eq_dh(b):
                return (b.dh ==
                        4 * b.spacer_porosity
                        / (2 / b.channel_height
                           + (1 - b.spacer_porosity) * 8 / b.channel_height))

        if self.config.concentration_polarization_type == ConcentrationPolarizationType.fixed:
            self.cp_modulus = Var(
                self.flowsheet().config.time,
                self.length_domain,
                solute_set,
                initialize=1.1,
                bounds=(0.9, 3),
                domain=NonNegativeReals,
                units=pyunits.dimensionless,
                doc='Concentration polarization modulus')

        # ==========================================================================
        # Volumetric Recovery rate
        @self.Constraint(self.flowsheet().config.time)
        def eq_recovery_vol_phase(b, t):
            return (b.recovery_vol_phase[t, 'Liq'] ==
                    b.mixed_permeate[t].flow_vol_phase['Liq'] /
                    b.feed_side.properties[t, self.first_element].flow_vol_phase['Liq'])

        # ==========================================================================
        # Mass-based Component Recovery rate
        @self.Constraint(self.flowsheet().config.time,
                         self.config.property_package.component_list)
        def eq_recovery_mass_phase_comp(b, t, j):
            return (b.recovery_mass_phase_comp[t, 'Liq', j]
                    * b.feed_side.properties[t, b.first_element].flow_mass_phase_comp['Liq', j] ==
                    b.mixed_permeate[t].flow_mass_phase_comp['Liq', j])

        @self.Constraint(self.flowsheet().config.time,
                         self.difference_elements,
                         self.config.property_package.phase_list,
                         self.config.property_package.component_list,
                         doc="Solvent and solute mass flux")
        def eq_flux_mass(b, t, x, p, j):
            prop_feed = b.feed_side.properties[t, x]
            prop_perm = b.permeate_side[t, x]
            interface = b.feed_side.properties_interface[t, x]
            comp = self.config.property_package.get_component(j)
            if comp.is_solvent():
                return (b.flux_mass_phase_comp[t, x, p, j] == b.A_comp[t, j] * b.dens_solvent
                        * ((prop_feed.pressure - prop_perm.pressure)
                           - (interface.pressure_osm - prop_perm.pressure_osm)))
            elif comp.is_solute():
                return (b.flux_mass_phase_comp[t, x, p, j] == b.B_comp[t, j]
                        * (interface.conc_mass_phase_comp[p, j] - prop_perm.conc_mass_phase_comp[p, j]))

        # # ==========================================================================
        # Concentration polarization
        @self.feed_side.Constraint(self.flowsheet().config.time,
                                   self.difference_elements,
                                   solute_set,
                                   doc="Concentration polarization")
        def eq_concentration_polarization(b, t, x, j):
            bulk = b.properties[t, x]
            interface = b.properties_interface[t, x]
            if self.config.concentration_polarization_type == ConcentrationPolarizationType.none:
                return interface.conc_mass_phase_comp['Liq', j] == \
                       bulk.conc_mass_phase_comp['Liq', j]
            elif self.config.concentration_polarization_type == ConcentrationPolarizationType.fixed:
                return (interface.conc_mass_phase_comp['Liq', j] ==
                        bulk.conc_mass_phase_comp['Liq', j]
                        * self.cp_modulus[t, x, j])
            elif self.config.concentration_polarization_type == ConcentrationPolarizationType.calculated:
                jw = self.flux_mass_phase_comp[t, x, 'Liq', 'H2O'] / self.dens_solvent
                js = self.flux_mass_phase_comp[t, x, 'Liq', j]
                return (interface.conc_mass_phase_comp['Liq', j] ==
                        bulk.conc_mass_phase_comp['Liq', j] * exp(jw / self.Kf[t, x, j])
                        - js / jw * (exp(jw / self.Kf[t, x, j]) - 1))

        # # ==========================================================================
        # Feed and permeate-side isothermal conditions
        @self.Constraint(self.flowsheet().config.time,
                         self.length_domain,
                         doc="Isothermal assumption for permeate")
        def eq_permeate_isothermal(b, t, x):
            return b.feed_side.properties[t, x].temperature == \
                   b.permeate_side[t, x].temperature

        # ==========================================================================
        # isothermal conditions at permeate outlet
        @self.Constraint(self.flowsheet().config.time,
                         doc="Isothermal assumption for permeate out")
        def eq_permeate_outlet_isothermal(b, t):
            return b.feed_side.properties[t, b.length_domain.last()].temperature == \
                   b.mixed_permeate[t].temperature

        # ==========================================================================
        # isobaric conditions across permeate channel and at permeate outlet
        @self.Constraint(self.flowsheet().config.time,
                         self.length_domain,
                         doc="Isobaric assumption for permeate out")
        def eq_permeate_outlet_isobaric(b, t, x):
            return b.permeate_side[t, x].pressure == \
                   b.mixed_permeate[t].pressure

        # rejection
        @self.Constraint(self.flowsheet().config.time,
                         solute_set)
        def eq_rejection_phase_comp(b, t, j):
            return (b.rejection_phase_comp[t, 'Liq', j] ==
                    1 - (b.mixed_permeate[t].conc_mass_phase_comp['Liq', j] /
                         b.feed_side.properties[t, self.first_element].conc_mass_phase_comp['Liq', j]))

        # ==========================================================================
        # Membrane area equation
        if hasattr(self, 'length') or hasattr(self, 'width'):
            @self.Constraint(doc="Membrane area")
            def eq_area(b):
                return b.area == b.length * b.width

        # # ==========================================================================
        # Constraints active when MassTransferCoefficient.calculated
        # Mass transfer coefficient calculation
        if self.config.mass_transfer_coefficient == MassTransferCoefficient.calculated:
            @self.Constraint(self.flowsheet().config.time,
                             self.difference_elements,
                             solute_set,
                             doc="Mass transfer coefficient in feed channel")
            def eq_Kf(b, t, x, j):
                bulk = b.feed_side.properties[t, x]
                return (b.Kf[t, x, j] * b.dh ==
                        bulk.diffus_phase['Liq']  # TODO: add diff coefficient to SW prop and consider multi-components
                        * b.N_Sh[t, x])

            @self.Constraint(self.flowsheet().config.time,
                             self.length_domain,
                             doc="Sherwood number")
            def eq_N_Sh(b, t, x):
                return (b.N_Sh[t, x] ==
                        0.46 * (b.N_Re[t, x] * b.N_Sc[t, x])**0.36)

            @self.Constraint(self.flowsheet().config.time,
                             self.length_domain,
                             doc="Schmidt number")
            def eq_N_Sc(b, t, x):
                bulk = b.feed_side.properties[t, x]
                return (b.N_Sc[t, x] * bulk.dens_mass_phase['Liq'] * bulk.diffus_phase['Liq'] ==
                        bulk.visc_d_phase['Liq'])

        if (self.config.mass_transfer_coefficient == MassTransferCoefficient.calculated
            or self.config.pressure_change_type == PressureChangeType.calculated):
            @self.Constraint(doc="Cross-sectional area")
            def eq_area_cross(b):
                return b.area_cross == b.channel_height * b.width * b.spacer_porosity

            @self.Constraint(self.flowsheet().config.time,
                             self.length_domain,
                             doc="Reynolds number")
            def eq_N_Re(b, t, x):
                bulk = b.feed_side.properties[t, x]
                return (b.N_Re[t, x] * b.area_cross * bulk.visc_d_phase['Liq'] ==
                        sum(bulk.flow_mass_phase_comp['Liq', j] for j in b.config.property_package.component_list)
                        * b.dh)

        if self.config.pressure_change_type == PressureChangeType.calculated:
            ## ==========================================================================
            # Crossflow velocity
            @self.Constraint(self.flowsheet().config.time,
                             self.length_domain,
                             doc="Crossflow velocity constraint")
            def eq_velocity(b, t, x):
                return b.velocity[t, x] * b.area_cross == b.feed_side.properties[t, x].flow_vol_phase['Liq']

            ## ==========================================================================
            # Darcy friction factor based on eq. S27 in SI for Cost Optimization of Osmotically Assisted Reverse Osmosis
            # TODO: this relationship for friction factor is specific to a particular spacer geometry. Add alternatives.
            @self.Constraint(self.flowsheet().config.time,
                             self.length_domain,
                             doc="Darcy friction factor constraint")
            def eq_friction_factor_darcy(b, t, x):
                return (b.friction_factor_darcy[t, x] - 0.42) * b.N_Re[t, x] == 189.3

            ## ==========================================================================
            # Pressure change per unit length due to friction
            # -1/2*f/dh*density*velocity^2
            @self.Constraint(self.flowsheet().config.time,
                             self.length_domain,
                             doc="pressure change per unit length due to friction")
            def eq_dP_dx(b, t, x):
                bulk = b.feed_side.properties[t, x]
                return (b.dP_dx[t, x] * b.dh ==
                        -0.5 * b.friction_factor_darcy[t, x]
                        * b.feed_side.properties[t, x].dens_mass_phase['Liq'] * b.velocity[t, x]**2)

        # ==========================================================================
        # Bulk and interface connections on the feed-side
        # TEMPERATURE
        @self.feed_side.Constraint(self.flowsheet().config.time,
                                   self.difference_elements,
                                   doc="Temperature at interface")
        def eq_equal_temp_interface(b, t, x):
            return b.properties_interface[t, x].temperature == \
                   b.properties[t, x].temperature
        # PRESSURE
        @self.feed_side.Constraint(self.flowsheet().config.time,
                                   self.difference_elements,
                                   doc="Pressure at interface")
        def eq_equal_pressure_interface(b, t, x):
            return b.properties_interface[t, x].pressure == \
                   b.properties[t, x].pressure
        # VOLUMETRIC FLOWRATE
        @self.feed_side.Constraint(self.flowsheet().config.time,
                                   self.difference_elements,
                                   doc="Volumetric flow at interface of inlet")
        def eq_equal_flow_vol_interface(b, t, x):
            return b.properties_interface[t, x].flow_vol_phase['Liq'] ==\
                   b.properties[t, x].flow_vol_phase['Liq']

        @self.Expression(self.flowsheet().config.time,
                         self.config.property_package.phase_list,
                         self.config.property_package.component_list,
                         doc="Average flux expression")
        def flux_mass_phase_comp_avg(b, t, p, j):
            return sum(b.flux_mass_phase_comp[t, x, p, j]
                       for x in self.difference_elements) / self.nfe

    def _add_expressions(self):
        """
        Generate expressions for additional results desired for full report
        """

        solute_set = self.config.property_package.solute_set
        if hasattr(self, 'N_Re'):
            @self.Expression(self.flowsheet().config.time,
                             doc="Average Reynolds Number expression")
            def N_Re_avg(b, t):
                return sum(b.N_Re[t, x]
                           for x in self.length_domain) / self.nfe
        if hasattr(self, 'Kf'):
            @self.Expression(self.flowsheet().config.time,
                             solute_set,
                             doc="Average mass transfer coefficient expression")
            def Kf_avg(b, t, j):
                return sum(b.Kf[t, x, j]
                           for x in self.difference_elements) / self.nfe

    def _get_stream_table_contents(self, time_point=0):
        return create_stream_table_dataframe(
            {
                "Feed Inlet": self.inlet,
                "Feed Outlet": self.retentate,
                "Permeate Outlet": self.permeate,
            },
            time_point=time_point,
        )

    def get_costing(self, module=None, **kwargs):
        self.costing = Block()
        module.ReverseOsmosis_costing(self.costing, **kwargs)

    def _get_state_args(self, source, mixed_permeate_properties, initialize_guess, state_args): 
        '''
        Arguments:
            source : property model containing inlet feed
            mixed_permeate_properties : mixed permeate property block
            initialize_guess : a dict of guesses for solvent_recovery, solute_recovery,
                               and cp_modulus. These guesses offset the initial values
                               for the retentate, permeate, and membrane interface
                               state blocks from the inlet feed
                               (default =
                               {'deltaP': -1e4,
                               'solvent_recovery': 0.5,
                               'solute_recovery': 0.01,
                               'cp_modulus': 1.1})
            state_args : a dict of arguments to be passed to the property
                         package(s) to provide an initial state for the inlet
                         feed side state block (see documentation of the specific
                         property package).
        '''

        # assumptions
        if initialize_guess is None:
            initialize_guess = {}
        if 'deltaP' not in initialize_guess:
            initialize_guess['deltaP'] = -1e4
        if 'solvent_recovery' not in initialize_guess:
            initialize_guess['solvent_recovery'] = 0.5
        if 'solute_recovery' not in initialize_guess:
            initialize_guess['solute_recovery'] = 0.01
        if 'cp_modulus' not in initialize_guess:
            initialize_guess['cp_modulus'] = 1.1

        if state_args is None:
            state_args = {}
            state_dict = source.define_port_members()

            for k in state_dict.keys():
                if state_dict[k].is_indexed():
                    state_args[k] = {}
                    for m in state_dict[k].keys():
                        state_args[k][m] = state_dict[k][m].value
                else:
                    state_args[k] = state_dict[k].value

        if 'flow_mass_phase_comp' not in state_args.keys():
            raise ConfigurationError(f'{self.__class__.__name__} initialization routine expects '
                                     'flow_mass_phase_comp as a state variable. Check '
                                     'that the property package supports this state '
                                     'variable or that the state_args provided to the '
                                     'initialize call includes this state variable')

        # slightly modify initial values for other state blocks
        state_args_retentate = deepcopy(state_args)
        state_args_permeate = deepcopy(state_args)

        state_args_retentate['pressure'] += initialize_guess['deltaP']
        state_args_permeate['pressure'] = mixed_permeate_properties.pressure.value
        for j in self.config.property_package.solvent_set:
            state_args_retentate['flow_mass_phase_comp'][('Liq', j)] *= (1 - initialize_guess['solvent_recovery'])
            state_args_permeate['flow_mass_phase_comp'][('Liq', j)] *= initialize_guess['solvent_recovery']
        for j in self.config.property_package.solute_set:
            state_args_retentate['flow_mass_phase_comp'][('Liq', j)] *= (1 - initialize_guess['solute_recovery'])
            state_args_permeate['flow_mass_phase_comp'][('Liq', j)] *= initialize_guess['solute_recovery']

        state_args_interface_in = deepcopy(state_args)
        state_args_interface_out = deepcopy(state_args_retentate)

        for j in self.config.property_package.solute_set:
            state_args_interface_in['flow_mass_phase_comp'][('Liq', j)] *= initialize_guess['cp_modulus']
            state_args_interface_out['flow_mass_phase_comp'][('Liq', j)] *= initialize_guess['cp_modulus']

        # TODO: I think this is what we'd like to do, but IDAES needs to be changed
        state_args_interface = {}
        for t in self.flowsheet().config.time:
            for x in self.length_domain:
                assert 0. <= x <= 1.
                state_args_tx = {}
                for k in state_args_interface_in:
                    if isinstance(state_args_interface_in[k], dict):
                        if k not in state_args_tx:
                            state_args_tx[k] = {}
                        for index in state_args_interface_in[k]:
                            state_args_tx[k][index] = (1.-x)*state_args_interface_in[k][index] + x*state_args_interface_out[k][index]
                    else:
                        state_args_tx[k] = (1.-x)*state_args_interface_in[k] + x*state_args_interface_out[k]
                state_args_interface[t,x] = state_args_tx

        x = 0.5
        state_args_tx = {}
        for k in state_args_interface_in:
            if isinstance(state_args_interface_in[k], dict):
                if k not in state_args_tx:
                    state_args_tx[k] = {}
                for index in state_args_interface_in[k]:
                    state_args_tx[k][index] = (1.-x)*state_args_interface_in[k][index] + x*state_args_interface_out[k][index]
            else:
                state_args_tx[k] = (1.-x)*state_args_interface_in[k] + x*state_args_interface_out[k]
        state_args_interface = state_args_tx

        return {'feed_side' : state_args,
                'retentate' : state_args_retentate,
                'permeate' : state_args_permeate,
                'interface' : state_args_interface,
               }

    def _get_performance_contents(self, time_point=0):
        x_in = self.first_element
        x_interface_in = self.difference_elements.first()
        x_out = self.length_domain.last()
        feed_inlet = self.feed_side.properties[time_point, x_in]
        feed_outlet = self.feed_side.properties[time_point, x_out]
        interface_inlet = self.feed_side.properties_interface[time_point, x_interface_in]
        interface_outlet = self.feed_side.properties_interface[time_point, x_out]
        permeate = self.mixed_permeate[time_point]
        var_dict = {}
        expr_dict = {}
        var_dict["Volumetric Recovery Rate"] = self.recovery_vol_phase[time_point, 'Liq']
        var_dict["Solvent Mass Recovery Rate"] = self.recovery_mass_phase_comp[time_point, 'Liq', 'H2O']
        var_dict["Membrane Area"] = self.area
        if hasattr(self, "length") and self.config.has_full_reporting:
            var_dict["Membrane Length"] = self.length
        if hasattr(self, "width") and self.config.has_full_reporting:
            var_dict["Membrane Width"] = self.width
        if hasattr(self, "deltaP") and self.config.has_full_reporting:
            var_dict["Pressure Change"] = self.deltaP[time_point]
        if hasattr(self, "N_Re") and self.config.has_full_reporting:
            var_dict["Reynolds Number @Inlet"] = self.N_Re[time_point, x_in]
            var_dict["Reynolds Number @Outlet"] = self.N_Re[time_point, x_out]
        if hasattr(self, "velocity") and self.config.has_full_reporting:
            var_dict["Velocity @Inlet"] = self.velocity[time_point, x_in]
            var_dict["Velocity @Outlet"] = self.velocity[time_point, x_out]
        for j in self.config.property_package.solute_set:
            if interface_inlet.is_property_constructed('conc_mass_phase_comp') and self.config.has_full_reporting:
                var_dict[f'{j} Concentration @Inlet,Membrane-Interface '] = (
                    interface_inlet.conc_mass_phase_comp['Liq', j])
            if interface_outlet.is_property_constructed('conc_mass_phase_comp') and self.config.has_full_reporting:
                var_dict[f'{j} Concentration @Outlet,Membrane-Interface '] = (
                    interface_outlet.conc_mass_phase_comp['Liq', j])
            if feed_inlet.is_property_constructed('conc_mass_phase_comp') and self.config.has_full_reporting:
                var_dict[f'{j} Concentration @Inlet,Bulk'] = (
                    feed_inlet.conc_mass_phase_comp['Liq', j])
            if feed_outlet.is_property_constructed('conc_mass_phase_comp') and self.config.has_full_reporting:
                var_dict[f'{j} Concentration @Outlet,Bulk'] = (
                    feed_outlet.conc_mass_phase_comp['Liq', j])
            if permeate.is_property_constructed('conc_mass_phase_comp') and self.config.has_full_reporting:
                var_dict[f'{j} Permeate Concentration'] = (
                    permeate.conc_mass_phase_comp['Liq', j])
        if interface_outlet.is_property_constructed('pressure_osm') and self.config.has_full_reporting:
            var_dict['Osmotic Pressure @Outlet,Membrane-Interface '] = (
                interface_outlet.pressure_osm)
        if feed_outlet.is_property_constructed('pressure_osm') and self.config.has_full_reporting:
            var_dict['Osmotic Pressure @Outlet,Bulk'] = (
                feed_outlet.pressure_osm)
        if interface_inlet.is_property_constructed('pressure_osm') and self.config.has_full_reporting:
            var_dict['Osmotic Pressure @Inlet,Membrane-Interface'] = (
                interface_inlet.pressure_osm)
        if feed_inlet.is_property_constructed('pressure_osm') and self.config.has_full_reporting:
            var_dict['Osmotic Pressure @Inlet,Bulk'] = (
                feed_inlet.pressure_osm)
        if feed_inlet.is_property_constructed('flow_vol_phase') and self.config.has_full_reporting:
            var_dict['Volumetric Flowrate @Inlet'] = (
                feed_inlet.flow_vol_phase['Liq'])
        if feed_outlet.is_property_constructed('flow_vol_phase') and self.config.has_full_reporting:
            var_dict['Volumetric Flowrate @Outlet'] = (
                feed_outlet.flow_vol_phase['Liq'])
        if hasattr(self, 'dh') and self.config.has_full_reporting:
            var_dict["Hydraulic Diameter"] = self.dh

        if self.config.has_full_reporting:
            expr_dict['Average Solvent Flux (LMH)'] = self.flux_mass_phase_comp_avg[time_point, 'Liq', 'H2O'] * 3.6e3
            expr_dict['Average Reynolds Number'] = self.N_Re_avg[time_point]
            for j in self.config.property_package.solute_set:
                expr_dict[f'{j} Average Solute Flux (GMH)'] = self.flux_mass_phase_comp_avg[time_point, 'Liq', j] * 3.6e6
                expr_dict[f'{j} Average Mass Transfer Coefficient (mm/h)'] = self.Kf_avg[time_point, j] * 3.6e6


        # TODO: add more vars
        return {"vars": var_dict, "exprs": expr_dict}

    # permeate properties need to rescale solute values by 100
    def _rescale_permeate_variable(self, var, factor=100):
        if var not in self._permeate_scaled_properties:
            sf = iscale.get_scaling_factor(var)
            iscale.set_scaling_factor(var, sf * factor)
            self._permeate_scaled_properties.add(var)

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        # these variables should have user input, if not there will be a warning
        if iscale.get_scaling_factor(self.area) is None:
            sf = iscale.get_scaling_factor(self.area, default=10, warning=True)
            iscale.set_scaling_factor(self.area, sf)

        if iscale.get_scaling_factor(self.A_comp) is None:
            iscale.set_scaling_factor(self.A_comp, 1e12)

        if iscale.get_scaling_factor(self.B_comp) is None:
            iscale.set_scaling_factor(self.B_comp, 1e8)

        if iscale.get_scaling_factor(self.recovery_vol_phase) is None:
            iscale.set_scaling_factor(self.recovery_vol_phase, 1)

        for (t, p, j), v in self.recovery_mass_phase_comp.items():
            if j in self.config.property_package.solvent_set:
                sf = 1
            elif j in self.config.property_package.solute_set:
                sf = 100
            if iscale.get_scaling_factor(v) is None:
                iscale.set_scaling_factor(v, sf)

        for v in self.rejection_phase_comp.values():
            if iscale.get_scaling_factor(v) is None:
                iscale.set_scaling_factor(v, 1)

        if hasattr(self, 'channel_height'):
            if iscale.get_scaling_factor(self.channel_height) is None:
                iscale.set_scaling_factor(self.channel_height, 1e3)

        if hasattr(self, 'spacer_porosity'):
            if iscale.get_scaling_factor(self.spacer_porosity) is None:
                iscale.set_scaling_factor(self.spacer_porosity, 1)

        if hasattr(self, 'dh'):
            if iscale.get_scaling_factor(self.dh) is None:
                iscale.set_scaling_factor(self.dh, 1e3)

        if hasattr(self, 'Kf'):
            for v in self.Kf.values():
                if iscale.get_scaling_factor(v) is None:
                    iscale.set_scaling_factor(v, 1e4)

        if hasattr(self, 'N_Re'):
            for t, x in self.N_Re.keys():
                if iscale.get_scaling_factor(self.N_Re[t, x]) is None:
                    iscale.set_scaling_factor(self.N_Re[t, x], 1e-2)

        if hasattr(self, 'N_Sc'):
            for t, x in self.N_Sc.keys():
                if iscale.get_scaling_factor(self.N_Sc[t, x]) is None:
                    iscale.set_scaling_factor(self.N_Sc[t, x], 1e-2)

        if hasattr(self, 'N_Sh'):
            for t, x in self.N_Sh.keys():
                if iscale.get_scaling_factor(self.N_Sh[t, x]) is None:
                     iscale.set_scaling_factor(self.N_Sh[t, x], 1e-2)

        if hasattr(self, 'velocity'):
            for v in self.velocity.values():
                if iscale.get_scaling_factor(v) is None:
                    iscale.set_scaling_factor(v, 1)

        if hasattr(self, 'friction_factor_darcy'):
            for v in self.friction_factor_darcy.values():
                if iscale.get_scaling_factor(v) is None:
                    iscale.set_scaling_factor(v, 1)

        if hasattr(self, 'cp_modulus'):
            for v in self.cp_modulus.values():
                if iscale.get_scaling_factor(v) is None:
                    iscale.set_scaling_factor(v, 1)

        for sb in (self.permeate_side, self.mixed_permeate):
            for blk in sb.values():
                for j in self.config.property_package.solute_set:
                    self._rescale_permeate_variable(blk.flow_mass_phase_comp['Liq', j])
                    if blk.is_property_constructed('mass_frac_phase_comp'):
                        self._rescale_permeate_variable(blk.mass_frac_phase_comp['Liq', j])
                    if blk.is_property_constructed('conc_mass_phase_comp'):
                        self._rescale_permeate_variable(blk.conc_mass_phase_comp['Liq', j])
                    if blk.is_property_constructed('mole_frac_phase_comp'):
                        self._rescale_permeate_variable(blk.mole_frac_phase_comp[j])
                    if blk.is_property_constructed('molality_comp'):
                        self._rescale_permeate_variable(blk.molality_comp[j])
                if blk.is_property_constructed('pressure_osm'):
                    self._rescale_permeate_variable(blk.pressure_osm)

        for (t, x, p, j), v in self.flux_mass_phase_comp.items():
            if iscale.get_scaling_factor(v) is None:
                comp = self.config.property_package.get_component(j)
                if comp.is_solvent():  # scaling based on solvent flux equation
                    sf = (iscale.get_scaling_factor(self.A_comp[t, j])
                          * iscale.get_scaling_factor(self.dens_solvent)
                          * iscale.get_scaling_factor(self.feed_side.properties[t, x].pressure))
                    iscale.set_scaling_factor(v, sf)
                elif comp.is_solute():  # scaling based on solute flux equation
                    sf = (iscale.get_scaling_factor(self.B_comp[t, j])
                          * iscale.get_scaling_factor(self.feed_side.properties[t, x].conc_mass_phase_comp[p, j]))
                    iscale.set_scaling_factor(v, sf)



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
This module contains a zero-order representation of a filter press unit
operation.
"""

from pyomo.environ import Constraint, units as pyunits, Var
from idaes.core import declare_process_block_class

from watertap.core import build_sido, ZeroOrderBaseData

# Some more information about this module
__author__ = "Kurban Sitterley"


@declare_process_block_class("FilterPressZO")
class FilterPressZOData(ZeroOrderBaseData):
    """
    Zero-Order model for a filter press unit operation.
    """

    CONFIG = ZeroOrderBaseData.CONFIG()

    def build(self):
        super().build()

        self._tech_type = "filter_press"

        build_sido(self)

        self.hours_per_day_operation = Var(
            self.flowsheet().time,
            units=pyunits.hour/pyunits.day,
            doc="Hours per day of filter press operation")

        self.cycle_time = Var(
            self.flowsheet().time,
            units=pyunits.hours,
            doc="Filter press cycle time")

        self.electricity_a_parameter = Var(
            self.flowsheet().time,
            units=pyunits.kWh/(pyunits.year*pyunits.ft**3),
            doc="Parameter A for electricity calculation")

        self.electricity_b_parameter = Var(
            self.flowsheet().time,
            units=pyunits.dimensionless,
            doc="Parameter B for electricity calculation")

        self._fixed_perf_vars.append(self.hours_per_day_operation)
        self._fixed_perf_vars.append(self.cycle_time)
        self._fixed_perf_vars.append(self.electricity_a_parameter)
        self._fixed_perf_vars.append(self.electricity_b_parameter)

        self.filter_press_capacity = Var(
            self.flowsheet().time,
            initialize=10,
            units=pyunits.ft**3,
            doc="Filter press capacity")
        
        self.electricity = Var(
            self.flowsheet().time,
            units=pyunits.kW,
            bounds=(0, None),
            doc="Filter press power")

        @self.Constraint(self.flowsheet().time,
                         doc="Filter press capacity constraint")
        def fp_capacity(b, t):
            Q = b.properties_in[t].flow_vol
            return (b.filter_press_capacity[t] == 
                pyunits.convert(Q, to_units=pyunits.ft**3/pyunits.day) /
                (b.hours_per_day_operation[t] / b.cycle_time[t]))
        
        @self.Constraint(self.flowsheet().time,
                         doc="Filter press electricity constraint")
        def fp_electricity(b, t):
            Q = b.properties_in[t].flow_vol
            A = pyunits.convert(
                    b.electricity_a_parameter[t] / 
                    (pyunits.kWh/(pyunits.year*pyunits.ft**3)),
                    to_units=pyunits.dimensionless)
            fp_cap = pyunits.convert(
                    b.filter_press_capacity[t] / pyunits.ft**3,
                    to_units=pyunits.dimensionless)
            return (b.electricity[t] == (A *
                    fp_cap ** b.electricity_b_parameter[t]) * 
                    (pyunits.kWh/pyunits.year) / 
                    pyunits.convert(Q, to_units=pyunits.m**3/pyunits.yr) * 
                    pyunits.convert(Q, to_units=pyunits.m**3/pyunits.hr))


        self._perf_var_dict["Filter Press Capacity (ft3)"] = self.filter_press_capacity
        self._perf_var_dict["Filter Press Power (kW)"] = self.electricity





from pyomo.environ import ConcreteModel, SolverFactory, TerminationCondition, \
    value, Constraint, Var, Objective, Expression
from pyomo.environ import units as pyunits
from pyomo.util.check_units import assert_units_consistent, assert_units_equivalent, check_units_equivalent
import pyomo.util.infeasible as infeas
from idaes.core import FlowsheetBlock
from idaes.core.util.model_statistics import degrees_of_freedom, number_variables, number_total_constraints, \
    number_activated_constraints, number_unfixed_variables_in_activated_equalities, number_activated_equalities, number_unused_variables

import idaes.core.util.model_statistics as stats
from idaes.core.util.constants import Constants
import idaes.core.util.scaling as iscale
import idaes.logger as idaeslog
from idaes.core.util import get_solver

from watertap.property_models import cryst_prop_pack as props
from watertap.unit_models.crystallizer import Crystallization

from io import StringIO
from pyomo.util.infeasible import (log_active_constraints, log_close_to_bounds,
                                   log_infeasible_bounds,
                                   log_infeasible_constraints)
from pyomo.common.log import LoggingIntercept
import logging


if __name__ == '__main__':

    # create model, flowsheet
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})
    # attach property package
    m.fs.properties = props.NaClParameterBlock()
    # build the unit model
    m.fs.crystallizer = Crystallization(default={
        "property_package": m.fs.properties})

    # now specify the model
    print('DOF before specifying:', degrees_of_freedom(m.fs))

    # Specify the Feed
    m.fs.crystallizer.inlet.flow_mass_phase_comp[0, 'Liq', 'NaCl'].fix(10.5119)
    m.fs.crystallizer.inlet.flow_mass_phase_comp[0, 'Liq', 'H2O'].fix(38.9326)
    # m.fs.crystallizer.properties_in[0].flow_vol_phase['Liq'].fix(150/3600)
    # m.fs.crystallizer.properties_in[0].mass_frac_phase_comp['Liq', 'NaCl'].fix(0.2126)
    m.fs.crystallizer.inlet.flow_mass_phase_comp[0, 'Sol', 'NaCl'].fix(1e-6)
    m.fs.crystallizer.inlet.flow_mass_phase_comp[0, 'Vap', 'H2O'].fix(1e-6)
    m.fs.crystallizer.inlet.pressure[0].fix(101325)
    m.fs.crystallizer.inlet.temperature[0].fix(273.15 + 20)
    print('DOF after specifying feed:', degrees_of_freedom(m.fs)) # Should be 2

    ##########################################
    # # Case 1: Fix crystallizer temperature
    ##########################################
    print("\n--- Case 1 ---")
    m.fs.crystallizer.temperature_operating.fix(273.15 + 55)
    #  m.fs.crystallizer.pressure_operating.fix(10000)
    m.fs.crystallizer.solids.flow_mass_phase_comp[0, 'Sol', 'NaCl'].fix(5.556)
    # m.fs.crystallizer.vapor.flow_mass_phase_comp[0, 'Vap', 'H2O'].fix(20.56)

    # Fix 
    m.fs.crystallizer.crystal_growth_rate.fix()
    m.fs.crystallizer.souders_brown_constant.fix()
    m.fs.crystallizer.crystal_median_length.fix()

    # # Scaling
    m.fs.properties.set_default_scaling('flow_mass_phase_comp', 1e-1, index=('Liq', 'H2O'))
    m.fs.properties.set_default_scaling('flow_mass_phase_comp', 1e-1, index=('Liq', 'NaCl'))
    m.fs.properties.set_default_scaling('flow_mass_phase_comp', 1e0, index=('Vap', 'H2O'))
    m.fs.properties.set_default_scaling('flow_mass_phase_comp', 1e-1, index=('Sol', 'NaCl'))
    iscale.calculate_scaling_factors(m.fs) 

    #  m.fs.crystallizer.k_param = 0.06
    # solving
    m.fs.crystallizer.initialize(outlvl=idaeslog.DEBUG)
    assert_units_consistent(m)  # check that units are consistent
    assert(degrees_of_freedom(m) == 0)  # check that the degrees of freedom are what we expect
    solver = get_solver()
    # solver = SolverFactory('ipopt')
    # solver.options = {'tol': 1e-8, 'nlp_scaling_method': 'user-scaling', 'halt_on_ampl_error': 'yes'}
    results = solver.solve(m, tee=True, symbolic_solver_labels=True)
    output = StringIO()
    with LoggingIntercept(output, 'pyomo.util.infeasible', logging.INFO):
        log_infeasible_constraints(m)
    print(output.getvalue().splitlines())
    assert results.solver.termination_condition == TerminationCondition.optimal
    # m.fs.crystallizer.display()

    m.fs.crystallizer.report()

    # assert False
    ##########################################
    # # Case 2: Fix crystallizer yield
    ##########################################
    print("\n--- Case 2 ---")
    # m.fs.crystallizer.T_crystallization.unfix()
    m.fs.crystallizer.solids.flow_mass_phase_comp[0, 'Sol', 'NaCl'].unfix()
    m.fs.crystallizer.crystallization_yield['NaCl'].fix(0.7)

    # solving
    assert_units_consistent(m)  # check that units are consistent
    assert(degrees_of_freedom(m) == 0)  # check that the degrees of freedom are what we expect
    # solver = SolverFactory('ipopt')
    # solver.options = {'tol': 1e-8, 'nlp_scaling_method': 'user-scaling'}
    solver = get_solver()
    results = solver.solve(m, tee=True)
    output = StringIO()
    with LoggingIntercept(output, 'pyomo.util.infeasible', logging.INFO):
        log_infeasible_constraints(m)
    print(output.getvalue().splitlines())
    assert results.solver.termination_condition == TerminationCondition.optimal
    # m.fs.crystallizer.display()

    m.fs.crystallizer.report()

    # ##########################################
    # # # Case 3: Fix crystallizer solids outlet
    # ##########################################
    print("\n--- Case 3 ---")
    m.fs.crystallizer.crystallization_yield['NaCl'].unfix()
    m.fs.crystallizer.product_volumetric_solids_fraction.fix(0.1182)

    # solving
    assert_units_consistent(m)  # check that units are consistent
    assert(degrees_of_freedom(m) == 0)  # check that the degrees of freedom are what we expect
    solver = get_solver()
    # solver = SolverFactory('ipopt')
    # solver.options = {'tol': 1e-8, 'nlp_scaling_method': 'user-scaling'}
    results = solver.solve(m, tee=True)
    output = StringIO()
    with LoggingIntercept(output, 'pyomo.util.infeasible', logging.INFO):
        log_infeasible_constraints(m)
    print(output.getvalue().splitlines())
    assert results.solver.termination_condition == TerminationCondition.optimal
    # m.fs.crystallizer.display()

    m.fs.crystallizer.report()

    #########################################
    # Case 4: Fix magma density
    #########################################
    print("\n--- Case 4 ---")
    m.fs.crystallizer.product_volumetric_solids_fraction.unfix()
    m.fs.crystallizer.dens_mass_magma.fix(250)

    # solving
    assert_units_consistent(m)  # check that units are consistent
    assert(degrees_of_freedom(m) == 0)  # check that the degrees of freedom are what we expect
    solver = get_solver()
    # solver = SolverFactory('ipopt')
    # solver.options = {'tol': 1e-8, 'nlp_scaling_method': 'user-scaling'}
    results = solver.solve(m, tee=True)
    output = StringIO()
    with LoggingIntercept(output, 'pyomo.util.infeasible', logging.INFO):
        log_infeasible_constraints(m)
    print(output.getvalue().splitlines())
    assert results.solver.termination_condition == TerminationCondition.optimal
    # m.fs.crystallizer.display()

    m.fs.crystallizer.report()

    #########################################
    # Case 5: Fix heat addition
    #########################################
    print("\n--- Case 5 ---")
    m.fs.crystallizer.dens_mass_magma.unfix()
    m.fs.crystallizer.work_mechanical[0].fix(55000)

    # solving
    assert_units_consistent(m)  # check that units are consistent
    assert(degrees_of_freedom(m) == 0)  # check that the degrees of freedom are what we expect
    solver = get_solver()
    # solver = SolverFactory('ipopt')
    # solver.options = {'tol': 1e-8, 'nlp_scaling_method': 'user-scaling'}
    results = solver.solve(m, tee=True)
    output = StringIO()
    with LoggingIntercept(output, 'pyomo.util.infeasible', logging.INFO):
        log_infeasible_constraints(m)
    print(output.getvalue().splitlines())
    assert results.solver.termination_condition == TerminationCondition.optimal
    # m.fs.crystallizer.display()

    m.fs.crystallizer.report()

    # print('number_variables:', number_variables(m.fs))
    # print('number_unused_variables:', number_unused_variables(m.fs))
    # print('number_total_constraints:', number_total_constraints(m.fs))
    # print('number_activated_constraints:', number_activated_constraints(m.fs))
    # print('number_activated_equalities', number_activated_equalities(m.fs))

